# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module containing a collection of plugins that produce data typically
found in Linux's /proc file system."""

import logging
from typing import Callable, Generator, Type, Optional

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class Maps(plugins.PluginInterface):
    """Lists all memory maps for all processes."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 3)

    MAXSIZE_DEFAULT = 1024 * 1024 * 1024  # 1 Gb

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed memory segments",
                default=False,
                optional=True,
            ),
            requirements.ListRequirement(
                name="address",
                description="Process virtual memory addresses to include "
                "(all other VMA sections are excluded). This can be any "
                "virtual address within the VMA section.",
                element_type=int,
                optional=True,
            ),
            requirements.IntRequirement(
                name="maxsize",
                description="Maximum size for dumped VMA sections "
                "(all the bigger sections will be ignored)",
                default=cls.MAXSIZE_DEFAULT,
                optional=True,
            ),
        ]

    @classmethod
    def list_vmas(
        cls,
        task: interfaces.objects.ObjectInterface,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            True
        ),
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Lists the Virtual Memory Areas of a specific process.

        Args:
            task: task object from which to list the vma
            filter_func: Function to take a vma and return False if it should be filtered out

        Returns:
            Yields vmas based on the task and filtered based on the filter function
        """
        mm_pointer = task.mm
        if not mm_pointer:
            vollog.debug(
                f"Excluded pid {task.pid} as there is no mm member. It is likely a kernel thread"
            )
            return

        if not mm_pointer.is_readable():
            vollog.error(f"Task {task.pid} has an invalid mm member")
            return

        for vma in mm_pointer.get_vma_iter():
            if filter_func(vma):
                yield vma
            else:
                vollog.debug(
                    f"Excluded vma at offset {vma.vol.offset:#x} for pid {task.pid} due to filter_func"
                )

    @classmethod
    def vma_dump(
        cls,
        context: interfaces.context.ContextInterface,
        task: interfaces.objects.ObjectInterface,
        vm_start: int,
        vm_end: int,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        maxsize: int = MAXSIZE_DEFAULT,
    ) -> Optional[interfaces.plugins.FileHandlerInterface]:
        """Extracts the complete data for VMA as a FileInterface.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            task: an task_struct instance
            vm_start: The start virtual address from the vma to dump
            vm_end: The end virtual address from the vma to dump
            open_method: class to provide context manager for opening the file
            maxsize: Max size of VMA section (default MAXSIZE_DEFAULT)

        Returns:
            An open FileInterface object containing the complete data for the task or None in the case of failure
        """
        pid = task.pid

        try:
            proc_layer_name = task.add_process_layer()
        except exceptions.InvalidAddressException as excp:
            vollog.debug(
                f"Process {pid}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
            )
            return None
        vm_size = vm_end - vm_start

        # check if vm_size is negative, this should never happen.
        if vm_size < 0:
            vollog.warning(
                f"Skip virtual memory dump for pid {pid} between {vm_start:#x}-{vm_end:#x} as {vm_size} is negative."
            )
            return None
        # check if vm_size is larger than the maxsize limit, and therefore is not saved out.
        if maxsize <= vm_size:
            vollog.warning(
                f"Skip virtual memory dump for pid {pid} between {vm_start:#x}-{vm_end:#x} as {vm_size} is larger than maxsize limit of {maxsize}"
            )
            return None
        proc_layer = context.layers[proc_layer_name]
        file_name = f"pid.{pid}.vma.{vm_start:#x}-{vm_end:#x}.dmp"
        try:
            file_handle = open_method(file_name)
            chunk_size = 1024 * 1024 * 10
            offset = vm_start
            while offset < vm_start + vm_size:
                to_read = min(chunk_size, vm_start + vm_size - offset)
                data = proc_layer.read(offset, to_read, pad=True)
                file_handle.write(data)
                offset += to_read
        except Exception as excp:
            vollog.debug(f"Unable to dump virtual memory {file_name}: {excp}")
            return None
        return file_handle

    def _generator(self, tasks):
        # build filter for addresses if required
        address_list = self.config.get("address", None)
        if not address_list:
            # do not filter as no address_list was supplied
            def vma_filter_func(_):
                return True

        else:
            # filter for any vm_start that matches the supplied address config
            def vma_filter_function(x: interfaces.objects.ObjectInterface) -> bool:
                addrs_in_vma = [
                    addr for addr in address_list if x.vm_start <= addr <= x.vm_end
                ]

                # if any of the user supplied addresses would fall within this vma return true
                return bool(addrs_in_vma)

            vma_filter_func = vma_filter_function

        for task in tasks:
            if not (task.mm and task.mm.is_readable()):
                continue
            name = utility.array_to_string(task.comm)

            for vma in self.list_vmas(task, filter_func=vma_filter_func):
                flags = vma.get_protection()
                page_offset = vma.get_page_offset()

                inode_num = None
                try:
                    dentry = vma.vm_file.get_dentry()
                    inode_ptr = dentry.d_inode
                    inode_num = inode_ptr.i_ino
                    major = inode_ptr.i_sb.major
                    minor = inode_ptr.i_sb.minor
                except exceptions.InvalidAddressException:
                    if not inode_num:
                        inode_num = 0
                    major = 0
                    minor = 0

                path = vma.get_name(self.context, task)

                file_output = "Disabled"
                if self.config["dump"]:
                    file_output = "Error outputting file"
                    try:
                        vm_start = vma.vm_start
                        vm_end = vma.vm_end
                    except AttributeError:
                        vollog.debug(
                            f"Unable to find the vm_start and vm_end for vma at {vma.vol.offset:#x} for pid {task.pid}"
                        )
                        vm_start = None
                        vm_end = None
                    if vm_start and vm_end:
                        # only attempt to dump the memory if we have vm_start and vm_end
                        file_handle = self.vma_dump(
                            self.context,
                            task,
                            vm_start,
                            vm_end,
                            self.open,
                            self.config["maxsize"],
                        )

                        if file_handle:
                            file_handle.close()
                            file_output = file_handle.preferred_filename
                yield (
                    0,
                    (
                        task.pid,
                        name,
                        format_hints.Hex(vma.vm_start),
                        format_hints.Hex(vma.vm_end),
                        flags,
                        format_hints.Hex(page_offset),
                        major,
                        minor,
                        inode_num,
                        path or renderers.NotAvailableValue(),
                        file_output,
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Start", format_hints.Hex),
                ("End", format_hints.Hex),
                ("Flags", str),
                ("PgOff", format_hints.Hex),
                ("Major", int),
                ("Minor", int),
                ("Inode", int),
                ("File Path", str),
                ("File output", str),
            ],
            self._generator(
                pslist.PsList.list_tasks(
                    self.context, self.config["kernel"], filter_func=filter_func
                )
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Volatility 3 - An open-source memory forensics framework"""

import inspect
import sys
from importlib import abc
from typing import List, TypeVar, Callable, Any, Optional

_T = TypeVar("_T")
_S = TypeVar("_S")


class classproperty(property):
    """Class property decorator.

    Note this will change the return type
    """

    def __init__(self, func: Callable[[_S], _T]) -> None:
        self._func = func
        super().__init__()

    def __get__(self, obj: Any, type: Optional[_S] = None) -> _T:
        if type is not None:
            return self._func(type)
        raise TypeError("Classproperty was not applied properly")


class WarningFindSpec(abc.MetaPathFinder):
    """Checks import attempts and throws a warning if the name shouldn't be
    used."""

    @staticmethod
    def find_spec(
        fullname: str, path: Optional[List[str]], target: None = None, **kwargs
    ) -> None:
        """Mock find_spec method that just checks the name, this must go
        first."""
        if fullname.startswith("volatility3.framework.plugins."):
            warning = f"Import {fullname}: Please do not use the volatility3.framework.plugins namespace directly, only use volatility3.plugins"
            # Pyinstaller uses walk_packages/_collect_submodules to import, but needs to read the modules to figure out dependencies
            # As such, we only print the warning when directly imported rather than from within walk_packages/_collect_submodules
            if inspect.stack()[-2].function not in [
                "walk_packages",
                "_collect_submodules",
            ] and inspect.stack()[-3].function not in ["_collect_submodules"]:
                raise Warning(warning)


warning_find_spec: List[abc.MetaPathFinder] = [WarningFindSpec()]
sys.meta_path = warning_find_spec + sys.meta_path

# We point the volatility3.plugins __path__ variable at BOTH
#   volatility3/plugins
#   volatility3/framework/plugins
# in that order.
#
# This will allow our users to override any component of any plugin without monkey patching,
# but it also allows us to clear out the plugins directory to get back to proper functionality.
# This offered the greatest flexibility for users whilst allowing us to keep the core separate and clean.
#
# This means that all plugins should be imported as volatility3.plugins (otherwise they'll be imported twice,
# once as volatility3.plugins.NAME and once as volatility3.framework.plugins.NAME).  We therefore throw an error
# if anyone tries to import anything under the volatility3.framework.plugins.* namespace
#
# The remediation is to only ever import form volatility3.plugins instead.



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

# Full details on the techniques used in these plugins to detect EDR-evading malware
# can be found in our 20 page whitepaper submitted to DEFCON along with the presentation
# https://www.volexity.com/wp-content/uploads/2024/08/Defcon24_EDR_Evasion_Detection_White-Paper_Andrew-Case.pdf

import logging

from typing import Tuple, Optional, Generator, List, Dict

from functools import partial

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
import volatility3.plugins.windows.pslist as pslist
import volatility3.plugins.windows.threads as threads
import volatility3.plugins.windows.pe_symbols as pe_symbols

vollog = logging.getLogger(__name__)


class DebugRegisters(interfaces.plugins.PluginInterface):
    # version 2.6.0 adds support for scanning for 'Ethread' structures by pool tags
    _required_framework_version = (2, 6, 0)
    _version = (1, 0, 1)

    @classmethod
    def get_requirements(cls) -> List:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="threads", component=threads.Threads, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pe_symbols", component=pe_symbols.PESymbols, version=(3, 0, 0)
            ),
        ]

    @staticmethod
    def _get_debug_info(
        ethread: interfaces.objects.ObjectInterface,
    ) -> Optional[Tuple[interfaces.objects.ObjectInterface, int, int, int, int, int]]:
        """
        Gathers information related to the debug registers for the given thread
        Args:
            ethread: the thread (_ETHREAD) to examine
        Returns:
            Tuple[interfaces.objects.ObjectInterface, int, int, int, int, int]: The owner process of the thread and the values for dr7, dr0, dr1, dr2, dr3
        """
        try:
            dr7 = ethread.Tcb.TrapFrame.Dr7
            state = ethread.Tcb.State
        except exceptions.InvalidAddressException:
            return None

        # 0 = debug registers not active
        # 4 = terminated
        if dr7 == 0 or state == 4:
            return None

        try:
            owner_proc = ethread.owning_process()
        except (AttributeError, exceptions.InvalidAddressException):
            return None

        dr0 = ethread.Tcb.TrapFrame.Dr0
        dr1 = ethread.Tcb.TrapFrame.Dr1
        dr2 = ethread.Tcb.TrapFrame.Dr2
        dr3 = ethread.Tcb.TrapFrame.Dr3

        # bail if all are 0
        if not (dr0 or dr1 or dr2 or dr3):
            return None

        return owner_proc, dr7, dr0, dr1, dr2, dr3

    def _generator(
        self,
    ) -> Generator[
        Tuple[
            int,
            Tuple[
                str,
                int,
                int,
                int,
                int,
                format_hints.Hex,
                str,
                str,
                format_hints.Hex,
                str,
                str,
                format_hints.Hex,
                str,
                str,
                format_hints.Hex,
                str,
                str,
            ],
        ],
        None,
        None,
    ]:
        vads_cache: Dict[int, pe_symbols.ranges_type] = {}

        proc_modules = None

        procs = pslist.PsList.list_processes(
            context=self.context, kernel_module_name=self.config["kernel"]
        )

        for proc in procs:
            for thread in threads.Threads.list_threads(
                self.context, self.config["kernel"], proc
            ):
                debug_info = self._get_debug_info(thread)
                if not debug_info:
                    continue

                owner_proc, dr7, dr0, dr1, dr2, dr3 = debug_info

                vads = pe_symbols.PESymbols.get_vads_for_process_cache(
                    vads_cache, owner_proc
                )
                if not vads:
                    continue

                # this lookup takes a while, so only perform if we need to
                if not proc_modules:
                    proc_modules = pe_symbols.PESymbols.get_process_modules(
                        self.context, self.config["kernel"], None
                    )
                    path_and_symbol = partial(
                        pe_symbols.PESymbols.path_and_symbol_for_address,
                        self.context,
                        self.config_path,
                        proc_modules,
                    )

                file0, sym0 = path_and_symbol(vads, dr0)
                file1, sym1 = path_and_symbol(vads, dr1)
                file2, sym2 = path_and_symbol(vads, dr2)
                file3, sym3 = path_and_symbol(vads, dr3)

                # if none map to an actual file VAD then bail
                if not (file0 or file1 or file2 or file3):
                    continue

                process_name = owner_proc.ImageFileName.cast(
                    "string",
                    max_length=owner_proc.ImageFileName.vol.count,
                    errors="replace",
                )

                thread_tid = thread.Cid.UniqueThread

                yield (
                    0,
                    (
                        process_name,
                        owner_proc.UniqueProcessId,
                        thread_tid,
                        thread.Tcb.State,
                        dr7,
                        format_hints.Hex(dr0),
                        file0 or renderers.NotApplicableValue(),
                        sym0 or renderers.NotApplicableValue(),
                        format_hints.Hex(dr1),
                        file1 or renderers.NotApplicableValue(),
                        sym1 or renderers.NotApplicableValue(),
                        format_hints.Hex(dr2),
                        file2 or renderers.NotApplicableValue(),
                        sym2 or renderers.NotApplicableValue(),
                        format_hints.Hex(dr3),
                        file3 or renderers.NotApplicableValue(),
                        sym3 or renderers.NotApplicableValue(),
                    ),
                )

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("Process", str),
                ("PID", int),
                ("TID", int),
                ("State", int),
                ("Dr7", int),
                ("Dr0", format_hints.Hex),
                ("Range0", str),
                ("Symbol0", str),
                ("Dr1", format_hints.Hex),
                ("Range1", str),
                ("Symbol1", str),
                ("Dr2", format_hints.Hex),
                ("Range2", str),
                ("Symbol2", str),
                ("Dr3", format_hints.Hex),
                ("Range3", str),
                ("Symbol3", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module for scanning translation layers looking for Windows PDB records
from loaded PE files.

This module contains a standalone scanner, and also a :class:`~volatility3.framework.interfaces.layers.ScannerInterface`
based scanner for use within the framework by calling :func:`~volatility3.framework.interfaces.layers.DataLayerInterface.scan`.
"""

import contextlib
import logging
import math
import os
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

from volatility3.framework import constants, exceptions, interfaces, layers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import intel, scanners
from volatility3.framework.symbols import native
from volatility3.framework.symbols.windows import pdbutil

if __name__ == "__main__":
    import sys

    sys.path.append(
        os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
    )

vollog = logging.getLogger(__name__)

ValidKernelType = Tuple[str, int, Dict[str, Optional[Union[bytes, str, int]]]]
KernelsType = Iterable[Dict[str, Any]]


class KernelPDBScanner(interfaces.automagic.AutomagicInterface):
    """Windows symbol loader based on PDB signatures.

    An Automagic object that looks for all Intel translation layers and scans each of them for a pdb signature.
    When found, a search for a corresponding Intermediate Format data file is carried out and if found an appropriate
    symbol space is automatically loaded.

    Once a specific kernel PDB signature has been found, a virtual address for the loaded kernel is determined
    by one of two methods.  The first method assumes a specific mapping from the kernel's physical address to its
    virtual address (typically the kernel is loaded at its physical location plus a specific offset).  The second method
    searches for a particular structure that lists the kernel module's virtual address, its size (not checked) and the
    module's name.  This value is then used if one was not found using the previous method.
    """

    priority = 30
    max_pdb_size = 0x400000
    exclusion_list = ["linux", "mac"]

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.VersionRequirement(
                name="pdb_utility",
                component=pdbutil.PDBUtility,
                version=(1, 0, 1),
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
        ]

    def find_virtual_layers_from_req(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        requirement: interfaces.configuration.RequirementInterface,
    ) -> List[str]:
        """Traverses the requirement tree, rooted at `requirement` looking for
        virtual layers that might contain a windows PDB.

        Returns a list of possible layers

        Args:
            context: The context in which the `requirement` lives
            config_path: The path within the `context` for the `requirement`'s configuration variables
            requirement: The root of the requirement tree to search for :class:~`volatility3.framework.interfaces.layers.TranslationLayerRequirement` objects to scan

        Returns:
            A list of (layer_name, scan_results)
        """
        sub_config_path = interfaces.configuration.path_join(
            config_path, requirement.name
        )
        results: List[str] = []
        if isinstance(requirement, requirements.TranslationLayerRequirement):
            # Check for symbols in this layer
            # FIXME: optionally allow a full (slow) scan
            # FIXME: Determine the physical layer no matter the virtual layer
            virtual_layer_name = context.config.get(sub_config_path, None)
            layer_name = context.config.get(
                interfaces.configuration.path_join(sub_config_path, "memory_layer"),
                None,
            )
            if layer_name and virtual_layer_name:
                memlayer = context.layers[virtual_layer_name]
                if isinstance(memlayer, intel.Intel):
                    results = [virtual_layer_name]
        else:
            for subreq in requirement.requirements.values():
                results += self.find_virtual_layers_from_req(
                    context, sub_config_path, subreq
                )
        return results

    def recurse_symbol_fulfiller(
        self,
        context: interfaces.context.ContextInterface,
        valid_kernel: ValidKernelType,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        """Fulfills the SymbolTableRequirements in `self._symbol_requirements`
        found by the `recurse_symbol_requirements`.

        This pass will construct any requirements that may need it in the context it was passed

        Args:
            context: Context on which to operate
            valid_kernel: A list of offsets where valid kernels have been found
            progress_callback: Means of providing the user with feedback during long processes
        """
        for sub_config_path, requirement in self._symbol_requirements:
            # TODO: Potentially think about multiple symbol requirements in both the same and different levels of the requirement tree
            # TODO: Consider whether a single found kernel can fulfill multiple requirements
            if valid_kernel:
                # TODO: Check that the symbols for this kernel will fulfill the requirement
                virtual_layer, _kvo, kernel = valid_kernel
                if not isinstance(kernel["pdb_name"], str) or not isinstance(
                    kernel["GUID"], str
                ):
                    raise TypeError("PDB name or GUID not a string value")

                pdbutil.PDBUtility.load_windows_symbol_table(
                    context=context,
                    guid=kernel["GUID"],
                    age=kernel["age"],
                    pdb_name=kernel["pdb_name"],
                    symbol_table_class="volatility3.framework.symbols.windows.WindowsKernelIntermedSymbols",
                    config_path=sub_config_path,
                    progress_callback=progress_callback,
                )
            else:
                vollog.debug("No suitable kernel pdb signature found")

    def set_kernel_virtual_offset(
        self,
        context: interfaces.context.ContextInterface,
        valid_kernel: ValidKernelType,
    ) -> None:
        """Traverses the requirement tree, looking for kernel_virtual_offset
        values that may need setting and sets it based on the previously
        identified `valid_kernel`.

        Args:
            context: Context on which to operate and provide the kernel virtual offset
            valid_kernel: List of valid kernels and offsets
        """
        if valid_kernel:
            # Set the virtual offset under the TranslationLayer it applies to
            virtual_layer, kvo, kernel = valid_kernel
            if kvo is not None:
                kvo_path = interfaces.configuration.path_join(
                    context.layers[virtual_layer].config_path, "kernel_virtual_offset"
                )
                context.config[kvo_path] = kvo
                vollog.debug(f"Setting kernel_virtual_offset to {hex(kvo)}")

    def get_physical_layer_name(self, context, vlayer):
        return context.config.get(
            interfaces.configuration.path_join(vlayer.config_path, "memory_layer"), None
        )

    def method_slow_scan(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        def test_virtual_kernel(
            physical_layer_name, virtual_layer_name: str, kernel: Dict[str, Any]
        ) -> Optional[ValidKernelType]:
            # It seems the kernel is loaded at a fixed mapping (presumably because the memory manager hasn't started yet)
            if kernel["mz_offset"] is None or not isinstance(kernel["mz_offset"], int):
                # Rule out kernels that couldn't find a suitable MZ header
                return None
            return (virtual_layer_name, kernel["mz_offset"], kernel)

        vollog.debug("Kernel base determination - optimized scan virtual layer")
        valid_kernel = self._method_layer_pdb_scan(
            context, vlayer, test_virtual_kernel, True, False, progress_callback
        )
        if valid_kernel is not None:
            return valid_kernel

        vollog.debug("Kernel base determination - slow scan virtual layer")
        return self._method_layer_pdb_scan(
            context, vlayer, test_virtual_kernel, False, False, progress_callback
        )

    def method_fixed_mapping(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        def test_physical_kernel(
            physical_layer_name: str, virtual_layer_name: str, kernel: Dict[str, Any]
        ) -> Optional[ValidKernelType]:
            # It seems the kernel is loaded at a fixed mapping (presumably because the memory manager hasn't started yet)
            if kernel["mz_offset"] is None or not isinstance(kernel["mz_offset"], int):
                # Rule out kernels that couldn't find a suitable MZ header
                return None
            if vlayer.bits_per_register == 64:
                kvo = kernel["mz_offset"] + (
                    31 << int(math.ceil(math.log2(vlayer.maximum_address + 1)) - 5)
                )
            else:
                kvo = kernel["mz_offset"] + (1 << (vlayer.bits_per_register - 1))
            try:
                kvp = vlayer.mapping(kvo, 0)
                if any(
                    (p == kernel["mz_offset"] and layer_name == physical_layer_name)
                    for (_, _, p, _, layer_name) in kvp
                ):
                    return (virtual_layer_name, kvo, kernel)
                else:
                    vollog.debug(
                        f"Potential kernel_virtual_offset did not map to expected location: {hex(kvo)}"
                    )
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Potential kernel_virtual_offset caused a page fault: {hex(kvo)}"
                )
            return None

        vollog.debug("Kernel base determination - testing fixed base address")
        return self._method_layer_pdb_scan(
            context, vlayer, test_physical_kernel, False, True, progress_callback
        )

    def _method_layer_pdb_scan(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        test_kernel: Callable,
        optimized: bool = False,
        physical: bool = True,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        # TODO: Verify this is a windows image
        valid_kernel = None
        virtual_layer_name = vlayer.name
        physical_layer_name = self.get_physical_layer_name(context, vlayer)

        layer_to_scan = physical_layer_name
        if not physical:
            layer_to_scan = virtual_layer_name

        start_scan_address = 0
        if (
            optimized
            and not physical
            and context.layers[layer_to_scan].metadata.architecture in ["Intel64"]
        ):
            # TODO: change this value accordingly when 5-Level paging is supported.
            start_scan_address = 0x1F0 << 39

        kernel_pdb_names = [
            bytes(name + ".pdb", "utf-8")
            for name in constants.windows.KERNEL_MODULE_NAMES
        ]
        kernels = pdbutil.PDBUtility.pdbname_scan(
            ctx=context,
            layer_name=layer_to_scan,
            start=start_scan_address,
            page_size=vlayer.page_size,
            pdb_names=kernel_pdb_names,
            progress_callback=progress_callback,
        )
        for kernel in kernels:
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Testing potential kernel for {kernel.get('pdb_name', 'Unknown')} at {kernel.get('signature_offset', -1):x} with MZ offset at {(kernel.get('mz_offset', -1) or -1):x}",
            )
            valid_kernel = test_kernel(physical_layer_name, virtual_layer_name, kernel)
            if valid_kernel is not None:
                break
        return valid_kernel

    def _method_offset(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        pattern: bytes,
        result_offset: int,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        """Method for finding a suitable kernel offset based on a module
        table."""
        vollog.debug(
            "Kernel base determination - searching layer module list structure"
        )
        valid_kernel: Optional[ValidKernelType] = None
        # If we're here, chances are high we're in a Win10 x64 image with kernel base randomization
        physical_layer_name = self.get_physical_layer_name(context, vlayer)
        physical_layer = context.layers[physical_layer_name]
        # TODO:  On older windows, this might be \WINDOWS\system32\nt rather than \SystemRoot\system32\nt
        results = physical_layer.scan(
            context, scanners.BytesScanner(pattern), progress_callback=progress_callback
        )
        seen: Set[int] = set()
        # Because this will launch a scan of the virtual layer, we want to be careful
        for result in results:
            # TODO: Identify the specific structure we're finding and document this a bit better
            pointer = context.object(
                "pdbscan!unsigned long long",
                offset=(result + result_offset),
                layer_name=physical_layer_name,
            )
            address = pointer & vlayer.address_mask
            if address in seen:
                continue
            seen.add(address)

            valid_kernel = self.check_kernel_offset(
                context, vlayer, address, progress_callback
            )

            if valid_kernel:
                break
        return valid_kernel

    def method_module_offset(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        return self._method_offset(
            context,
            vlayer,
            b"\\SystemRoot\\system32\\nt",
            -16 - int(vlayer.bits_per_register / 8),
            progress_callback,
        )

    def method_kdbg_offset(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        return self._method_offset(context, vlayer, b"KDBG", 8, progress_callback)

    def check_kernel_offset(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        address: int,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        """Scans a virtual address."""
        # Scan a few megs of the virtual space at the location to see if they're potential kernels

        valid_kernel: Optional[ValidKernelType] = None
        kernel_pdb_names = [
            bytes(name + ".pdb", "utf-8")
            for name in constants.windows.KERNEL_MODULE_NAMES
        ]

        virtual_layer_name = vlayer.name
        with contextlib.suppress(exceptions.InvalidAddressException):
            if vlayer.read(address, 0x2) == b"MZ":
                res = list(
                    pdbutil.PDBUtility.pdbname_scan(
                        ctx=context,
                        layer_name=vlayer.name,
                        page_size=vlayer.page_size,
                        pdb_names=kernel_pdb_names,
                        progress_callback=progress_callback,
                        start=address,
                        end=address + self.max_pdb_size,
                    )
                )
                if res:
                    valid_kernel = (virtual_layer_name, address, res[0])
        return valid_kernel

    def method_low_stub_offset(
        self,
        context: interfaces.context.ContextInterface,
        vlayer: layers.intel.Intel,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        # This method is only valid for x64 systems
        if not isinstance(vlayer, intel.Intel32e):
            return None
        kernel_hint = 0
        kernel_base = 0
        physical_layer = context.layers.get("memory_layer")

        # Try locating kernel base via x64 Low Stub in lower 1MB starting from second page (4KB)
        # If "Discard Low Memory" setting is disabled in BIOS, the Low Stub may be at the third/fourth or further pages
        for offset in range(0x1000, 0x100000, 0x1000):
            try:
                jmp_and_completion_values = int.from_bytes(
                    physical_layer.read(offset, 0x8), "little"
                )
                if (
                    0xFFFFFFFFFFFF00FF & jmp_and_completion_values
                    != constants.windows.JMP_AND_COMPLETION_SIGNATURE
                ):
                    continue
                cr3_value = int.from_bytes(
                    physical_layer.read(
                        offset + constants.windows.PROCESSOR_START_BLOCK_CR3_OFFSET, 0x8
                    ),
                    "little",
                )

                # Compare previously observed valid page table address that's stored in vlayer._initial_entry
                # with PROCESSOR_START_BLOCK->ProcessorState->SpecialRegisters->Cr3
                # which was observed to be an invalid page address, so add 1 (to make it valid too)
                if (cr3_value + 1) != vlayer._initial_entry:
                    continue
                potential_kernel_hint = int.from_bytes(
                    physical_layer.read(
                        offset
                        + constants.windows.PROCESSOR_START_BLOCK_LM_TARGET_OFFSET,
                        0x8,
                    ),
                    "little",
                )
                if 0x3 & potential_kernel_hint:
                    continue
                kernel_hint = potential_kernel_hint & 0xFFFFFFFFFFFF
                kernel_base = kernel_hint & (~0x1FFFFF) & 0xFFFFFFFFFFFF
                break
            except exceptions.InvalidAddressException:
                continue

        if kernel_base:
            # Scanning 32mb in 2mb chunks for the 'ntoskrnl' base address
            while (kernel_base + 0x2000000) > kernel_hint:
                for i in range(0, 0x200000, 0x1000):
                    valid_kernel = self.check_kernel_offset(
                        context, vlayer, kernel_base + i, progress_callback
                    )
                    if valid_kernel:
                        return valid_kernel
                kernel_base -= 0x200000
        return None

    # List of methods to be run, in order, to determine the valid kernels
    methods = [
        method_low_stub_offset,
        method_kdbg_offset,
        method_module_offset,
        method_fixed_mapping,
        method_slow_scan,
    ]

    def determine_valid_kernel(
        self,
        context: interfaces.context.ContextInterface,
        potential_layers: List[str],
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[ValidKernelType]:
        """Runs through the identified potential kernels and verifies their
        suitability.

        This carries out a scan using the pdb_signature scanner on a physical layer.  It uses the
        results of the scan to determine the virtual offset of the kernel.  On early windows implementations
        there is a fixed mapping between the physical and virtual addresses of the kernel.  On more recent versions
        a search is conducted for a structure that will identify the kernel's virtual offset.

        Args:
            context: Context on which to operate
            potential_layers: List of layer names that the kernel might live at
            progress_callback: Function taking a percentage and optional description to be called during expensive computations to indicate progress

        Returns:
            A dictionary of valid kernels
        """
        valid_kernel: Optional[ValidKernelType] = None
        for virtual_layer_name in potential_layers:
            vlayer = context.layers.get(virtual_layer_name, None)
            if isinstance(vlayer, layers.intel.Intel):
                for method in self.methods:
                    valid_kernel = method(self, context, vlayer, progress_callback)
                    if valid_kernel:
                        break
        if not valid_kernel:
            vollog.info("No suitable kernels found during pdbscan")
        return valid_kernel

    def __call__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        requirement: interfaces.configuration.RequirementInterface,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        if requirement.unsatisfied(context, config_path):
            if "pdbscan" not in context.symbol_space:
                context.symbol_space.append(
                    native.NativeTable("pdbscan", native.std_ctypes)
                )
            # TODO: check if this is a windows symbol requirement, otherwise ignore it
            self._symbol_requirements = self.find_requirements(
                context, config_path, requirement, requirements.SymbolTableRequirement
            )
            potential_layers = self.find_virtual_layers_from_req(
                context=context, config_path=config_path, requirement=requirement
            )
            for sub_config_path, symbol_req in self._symbol_requirements:
                parent_path = interfaces.configuration.parent_path(sub_config_path)
                if symbol_req.unsatisfied(context, parent_path):
                    valid_kernel = self.determine_valid_kernel(
                        context, potential_layers, progress_callback
                    )
                    if valid_kernel:
                        self.set_kernel_virtual_offset(context, valid_kernel)
                        self.recurse_symbol_fulfiller(
                            context, valid_kernel, progress_callback
                        )

        if progress_callback is not None:
            progress_callback(100, "PDB scanning finished")



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import interfaces, renderers, exceptions, constants
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import linux

vollog = logging.getLogger(__name__)


class Tty_Check(plugins.PluginInterface):
    """Checks tty devices for hooks"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherers",
                component=linux_utilities_modules.ModuleGatherers,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 0, 0)
            ),
        ]

    def _generator(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        try:
            tty_drivers = vmlinux.object_from_symbol("tty_drivers").cast("list_head")
        except exceptions.SymbolError:
            tty_drivers = None

        if not tty_drivers:
            raise TypeError(
                "This plugin requires the tty_drivers structure."
                "This structure is not present in the supplied symbol table."
                "This means you are either analyzing an unsupported kernel version or that your symbol table is corrupt."
            )

        known_modules = linux_utilities_modules.Modules.run_modules_scanners(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            caller_wanted_gatherers=linux_utilities_modules.ModuleGatherers.all_gatherers_identifier,
        )

        for tty in tty_drivers.to_list(
            vmlinux.symbol_table_name + constants.BANG + "tty_driver", "tty_drivers"
        ):
            try:
                ttys = utility.array_of_pointers(
                    tty.ttys.dereference(),
                    count=tty.num,
                    subtype=vmlinux.symbol_table_name + constants.BANG + "tty_struct",
                    context=self.context,
                )
            except exceptions.PagedInvalidAddressException:
                continue

            for tty_dev in ttys:
                if tty_dev == 0:
                    continue

                try:
                    name = utility.array_to_string(tty_dev.name)
                    recv_buf = tty_dev.ldisc.ops.receive_buf
                except exceptions.InvalidAddressException:
                    continue

                module_info, symbol_name = (
                    linux_utilities_modules.Modules.module_lookup_by_address(
                        self.context, vmlinux.name, known_modules, recv_buf
                    )
                )

                if module_info:
                    module_name = module_info.name
                else:
                    module_name = renderers.NotAvailableValue()

                yield (
                    0,
                    (
                        name,
                        format_hints.Hex(recv_buf),
                        module_name,
                        symbol_name or renderers.NotAvailableValue(),
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Name", str),
                ("Address", format_hints.Hex),
                ("Module", str),
                ("Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import interfaces, renderers
from volatility3.framework.renderers import format_hints
from volatility3.framework.configuration import requirements
from volatility3.plugins.linux import pslist


class Check_creds(interfaces.plugins.PluginInterface):
    """Checks if any processes are sharing credential structures"""

    _required_framework_version = (2, 0, 0)
    _version = (2, 0, 2)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
        ]

    def _generator(self):
        vmlinux = self.context.modules[self.config["kernel"]]

        type_task = vmlinux.get_type("task_struct")

        if not type_task.has_member("cred"):
            raise TypeError(
                "This plugin requires the task_struct structure to have a cred member. "
                "This member is not present in the supplied symbol table. "
                "This means you are either analyzing an unsupported kernel version or that your symbol table is corrupt."
            )

        creds = {}

        tasks = pslist.PsList.list_tasks(self.context, vmlinux.name)

        for task in tasks:
            task_cred_ptr = task.cred
            if not (task_cred_ptr and task_cred_ptr.is_readable()):
                continue

            cred_addr = task_cred_ptr.dereference().vol.offset

            creds.setdefault(cred_addr, [])
            creds[cred_addr].append(task.pid)

        for cred_addr, pids in creds.items():
            if len(pids) > 1:
                pid_str = ", ".join(str(pid) for pid in pids)

                fields = [
                    format_hints.Hex(cred_addr),
                    pid_str,
                ]
                yield (0, fields)

    def run(self):
        headers = [
            ("CredVAddr", format_hints.Hex),
            ("PIDs", str),
        ]
        return renderers.TreeGrid(headers, self._generator())



# This file is Copyright 2021 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import datetime
from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints, conversion
from volatility3.framework.objects import utility
from volatility3.framework.layers import crash

vollog = logging.getLogger(__name__)


class Crashinfo(interfaces.plugins.PluginInterface):
    """Lists the information from a Windows crash dump."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.TranslationLayerRequirement(
                name="primary",
                description="Memory layer for the kernel",
                architectures=["Intel32", "Intel64"],
            ),
        ]

    def _generator(self, layer: crash.WindowsCrashDump32Layer):
        header = layer.get_header()
        uptime = datetime.timedelta(microseconds=int(header.SystemUpTime) / 10)

        if header.DumpType == 0x1:
            dump_type = "Full Dump (0x1)"
        elif header.DumpType == 0x5:
            dump_type = "Bitmap Dump (0x5)"
        else:
            # this should never happen since the crash layer only accepts 0x1 and 0x5
            dump_type = f"Unknown/Unsupported ({header.DumpType:#x})"

        if header.DumpType == 0x5:
            summary_header = layer.get_summary_header()
            bitmap_header_size = format_hints.Hex(summary_header.HeaderSize)
            bitmap_size = format_hints.Hex(summary_header.BitmapSize)
            bitmap_pages = format_hints.Hex(summary_header.Pages)
        else:
            bitmap_header_size = bitmap_size = bitmap_pages = (
                renderers.NotApplicableValue()
            )

        yield (
            0,
            (
                utility.array_to_string(header.Signature),
                header.MajorVersion,
                header.MinorVersion,
                format_hints.Hex(header.DirectoryTableBase),
                format_hints.Hex(header.PfnDataBase),
                format_hints.Hex(header.PsLoadedModuleList),
                format_hints.Hex(header.PsActiveProcessHead),
                header.MachineImageType,
                header.NumberProcessors,
                format_hints.Hex(header.KdDebuggerDataBlock),
                dump_type,
                str(uptime),
                utility.array_to_string(header.Comment),
                conversion.wintime_to_datetime(header.SystemTime),
                bitmap_header_size,
                bitmap_size,
                bitmap_pages,
            ),
        )

    def run(self):
        crash_layer = None
        for layer_name in self._context.layers:
            layer = self._context.layers[layer_name]
            if isinstance(layer, crash.WindowsCrashDump32Layer):
                crash_layer = layer
                break

        if crash_layer is None:
            vollog.error("This plugin requires a Windows crash dump")
            raise

        return renderers.TreeGrid(
            [
                ("Signature", str),
                ("MajorVersion", int),
                ("MinorVersion", int),
                ("DirectoryTableBase", format_hints.Hex),
                ("PfnDataBase", format_hints.Hex),
                ("PsLoadedModuleList", format_hints.Hex),
                ("PsActiveProcessHead", format_hints.Hex),
                ("MachineImageType", int),
                ("NumberProcessors", int),
                ("KdDebuggerDataBlock", format_hints.Hex),
                ("DumpType", str),
                ("SystemUpTime", str),
                ("Comment", str),
                ("SystemTime", datetime.datetime),
                ("BitmapHeaderSize", format_hints.Hex),
                ("BitmapSize", format_hints.Hex),
                ("BitmapPages", format_hints.Hex),
            ],
            self._generator(crash_layer),
        )



# We use the SemVer 2.0.0 versioning scheme
VERSION_MAJOR = 2  # Number of releases of the library with a breaking change
VERSION_MINOR = 28  # Number of changes that only add to the interface
VERSION_PATCH = 0  # Number of changes that do not change the interface
VERSION_SUFFIX = ""

PACKAGE_VERSION = (
    ".".join(str(x) for x in [VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH])
    + VERSION_SUFFIX
)
"""The canonical version of the volatility3 package"""



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import NamedTuple, Dict, Generator

from volatility3.framework import interfaces, exceptions, constants
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.plugins.windows import pslist, vadinfo

vollog = logging.getLogger(__name__)


class VadData(NamedTuple):
    protection: str
    path: str


class DLLData(NamedTuple):
    path: str


### Useful references on process hollowing
# https://cysinfo.com/detecting-deceptive-hollowing-techniques/
# https://github.com/m0n0ph1/Process-Hollowing


class HollowProcesses(interfaces.plugins.PluginInterface):
    """Lists hollowed processes"""

    _required_framework_version = (2, 4, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="vadinfo", component=vadinfo.VadInfo, version=(2, 0, 0)
            ),
        ]

    def _get_vads_data(
        self, proc: interfaces.objects.ObjectInterface
    ) -> Dict[int, VadData]:
        """
        Returns a dictionary of:
            base address -> (protection string, file name)
        For each mapped VAD in the process. This is used
        for quick lookups of data and matching the DLL
        at the same base address as the VAD
        """
        vads = {}

        kernel = self.context.modules[self.config["kernel"]]

        for vad in proc.get_vad_root().traverse():
            protection_string = vad.get_protection(
                vadinfo.VadInfo.protect_values(
                    self.context, kernel.layer_name, kernel.symbol_table_name
                ),
                vadinfo.winnt_protections,
            )

            fn = vad.get_file_name()
            if not fn or not isinstance(fn, str):
                fn = "<Non-File Backed Region>"

            vads[vad.get_start()] = VadData(protection_string, fn)

        return vads

    def _get_dlls_map(
        self, proc: interfaces.objects.ObjectInterface
    ) -> Dict[int, DLLData]:
        """
        Returns a dictionary of:
            base address -> path
        for each DLL loaded in the process

        This is used to cross compare with
        the corresponding VAD and to have a
        backup path source in case of smear
        in the VAD
        """
        dlls = {}

        for entry in proc.load_order_modules():
            try:
                base = entry.DllBase
            except exceptions.InvalidAddressException:
                continue

            try:
                FullDllName = entry.FullDllName.get_string()
            except exceptions.InvalidAddressException:
                FullDllName = renderers.UnreadableValue()

            dlls[base] = DLLData(FullDllName)

        return dlls

    def _get_image_base(self, proc: interfaces.objects.ObjectInterface) -> int:
        """
        Uses the PEB to get the image base of the process
        """
        kernel = self.context.modules[self.config["kernel"]]

        try:
            proc_layer_name = proc.add_process_layer()
            peb = self.context.object(
                kernel.symbol_table_name + constants.BANG + "_PEB",
                layer_name=proc_layer_name,
                offset=proc.Peb,
            )
            return peb.ImageBaseAddress
        except exceptions.InvalidAddressException:
            return None

    def _check_load_address(self, proc, _, __) -> Generator[str, None, None]:
        """
        Detects when the image base in the PEB, which is writable by process malware,
        does not match the section base address - whose value lives in kernel memory.
        Many malware samples will manipulate their image base to fool AVs/EDRs and
        as a necessary part of certain hollowing techniques
        """
        image_base = self._get_image_base(proc)
        if image_base is not None and image_base != proc.SectionBaseAddress:
            yield f"The ImageBaseAddress reported from the PEB ({image_base:#x}) does not match the process SectionBaseAddress ({proc.SectionBaseAddress:#x})"

    def _check_exe_protection(
        self, proc, vads: Dict[int, VadData], __
    ) -> Generator[str, None, None]:
        """
        Legitimately mapped application executables and DLLs
        will have a VAD present and its initial protection will be
        PAGE_EXECUTE_WRITECOPY.
        Many process hollowing and code injection techniques will
        unmap the real executable and/or map in executables with
        incorrect permissions.
        This check verifies the VAD for the application exe.
        `_check_dlls_protection` checks for DLLs mapped in the process.
        """
        base = proc.SectionBaseAddress

        if base not in vads:
            yield f"There is no VAD starting at the base address of the process executable ({base:#x})"
        elif vads[base].protection != "PAGE_EXECUTE_WRITECOPY":
            yield f"Unexpected protection ({vads[base].protection}) for VAD hosting the process executable ({base:#x}) with path {vads[base].path}"

    def _check_dlls_protection(
        self, _, vads: Dict[int, VadData], dlls: Dict[int, DLLData]
    ) -> Generator[str, None, None]:
        for dll_base in dlls:
            # could be malicious but triggers too many FPs from smear
            if dll_base not in vads:
                continue

            # PAGE_EXECUTE_WRITECOPY is the only valid permission for mapped DLLs and .exe files
            if vads[dll_base].protection != "PAGE_EXECUTE_WRITECOPY":
                yield f"Unexpected protection ({vads[dll_base].protection}) for DLL in the PEB's load order list ({dll_base:#x}) with path {dlls[dll_base].path}"

    def _generator(self, procs):
        checks = [
            self._check_load_address,
            self._check_exe_protection,
            self._check_dlls_protection,
        ]

        for proc in procs:
            # smear and/or terminated process
            dlls = self._get_dlls_map(proc)
            if len(dlls) < 3:
                continue

            vads = self._get_vads_data(proc)
            if len(vads) < 5:
                continue

            proc_name = utility.array_to_string(proc.ImageFileName)
            pid = proc.UniqueProcessId

            for check in checks:
                for note in check(proc, vads, dlls):
                    yield (
                        0,
                        (
                            pid,
                            proc_name,
                            note,
                        ),
                    )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Notes", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import datetime
import logging
from typing import Callable, Dict, Set, Tuple

from volatility3.framework import interfaces, renderers, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class PsTree(interfaces.plugins.PluginInterface):
    """Plugin for listing processes in a tree based on their parent process ID."""

    _required_framework_version = (2, 0, 0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._processes: Dict[int, Tuple[interfaces.objects.ObjectInterface, int]] = {}
        self._levels: Dict[int, int] = {}
        self._children: Dict[int, Set[int]] = {}
        self._ancestors: Set[int] = set([])

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.BooleanRequirement(
                name="physical",
                description="Display physical offsets instead of virtual",
                default=pslist.PsList.PHYSICAL_DEFAULT,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process ID to include (with ancestors and descendants, all other processes are excluded)",
                optional=True,
            ),
        ]

    def find_level(
        self,
        pid: int,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            False
        ),
    ) -> None:
        """Finds how deep the pid is in the processes list."""
        seen = {pid}
        level = 0
        proc, _ = self._processes.get(pid, None)
        filtered = not filter_func(proc)
        while proc is not None and proc.InheritedFromUniqueProcessId not in seen:
            if filtered:
                self._ancestors.add(proc.UniqueProcessId)
            child_list = self._children.get(proc.InheritedFromUniqueProcessId, set([]))
            child_list.add(proc.UniqueProcessId)
            self._children[proc.InheritedFromUniqueProcessId] = child_list
            seen.add(proc.InheritedFromUniqueProcessId)
            proc, _ = self._processes.get(
                proc.InheritedFromUniqueProcessId, (None, None)
            )
            level += 1
        self._levels[pid] = level

    def _generator(
        self,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            False
        ),
    ):
        """Generates the Tree of processes."""
        kernel = self.context.modules[self.config["kernel"]]

        for proc in pslist.PsList.list_processes(
            context=self.context, kernel_module_name=self.config["kernel"]
        ):
            if not self.config.get("physical", pslist.PsList.PHYSICAL_DEFAULT):
                offset = proc.vol.offset
            else:
                layer_name = kernel.layer_name
                memory = self.context.layers[layer_name]
                (_, _, offset, _, _) = list(
                    memory.mapping(offset=proc.vol.offset, length=0)
                )[0]

            self._processes[proc.UniqueProcessId] = proc, offset

        # Build the child/level maps
        for pid in self._processes:
            self.find_level(pid, filter_func)

        process_pids = set([])

        def yield_processes(pid: int, descendant: bool = False):
            if pid in process_pids:
                vollog.debug(f"Pid cycle: already processed pid {pid}")
                return None

            process_pids.add(pid)

            if pid not in self._ancestors and not descendant:
                vollog.debug(f"Pid cycle: pid {pid} not in filtered tree")
                return None

            proc, offset = self._processes[pid]
            row = (
                proc.UniqueProcessId,
                proc.InheritedFromUniqueProcessId,
                proc.ImageFileName.cast(
                    "string", max_length=proc.ImageFileName.vol.count, errors="replace"
                ),
                format_hints.Hex(offset),
                proc.ActiveThreads,
                proc.get_handle_count(),
                proc.get_session_id(),
                proc.get_is_wow64(),
                proc.get_create_time(),
                proc.get_exit_time(),
            )

            try:
                audit = proc.SeAuditProcessCreationInfo.ImageFileName.Name
                # If 'audit' is set to the empty string, display NotAvailableValue
                row += (audit.get_string() or renderers.NotAvailableValue(),)
            except exceptions.InvalidAddressException:
                row += (renderers.NotAvailableValue(),)

            try:
                process_params = proc.get_peb().ProcessParameters
                row += (
                    process_params.CommandLine.get_string(),
                    process_params.ImagePathName.get_string(),
                )
            except exceptions.InvalidAddressException:
                row += (
                    renderers.NotAvailableValue(),
                    renderers.NotAvailableValue(),
                )

            yield (self._levels[pid] - 1, row)
            for child_pid in self._children.get(pid, []):
                yield from yield_processes(
                    child_pid, descendant or not filter_func(proc)
                )

        for pid in self._levels:
            if self._levels[pid] == 1:
                yield from yield_processes(pid)

    def run(self):
        offsettype = (
            "(V)"
            if not self.config.get("physical", pslist.PsList.PHYSICAL_DEFAULT)
            else "(P)"
        )

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("PPID", int),
                ("ImageFileName", str),
                (f"Offset{offsettype}", format_hints.Hex),
                ("Threads", int),
                ("Handles", int),
                ("SessionId", int),
                ("Wow64", bool),
                ("CreateTime", datetime.datetime),
                ("ExitTime", datetime.datetime),
                ("Audit", str),
                ("Cmd", str),
                ("Path", str),
            ],
            self._generator(
                filter_func=pslist.PsList.create_pid_filter(
                    self.config.get("pid", None)
                ),
            ),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.renderers import format_hints
from volatility3.framework.interfaces import plugins
from volatility3.framework.configuration import requirements

vollog = logging.getLogger(__name__)


class EBPF(plugins.PluginInterface):
    """Enumerate eBPF programs"""

    _required_framework_version = (2, 0, 0)

    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
        ]

    def get_ebpf_programs(
        self,
        context: interfaces.context.ContextInterface,
        vmlinux_module_name: str,
    ) -> interfaces.objects.ObjectInterface:
        """Enumerate eBPF programs walking its IDR.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            vmlinux_module_name: The name of the kernel module on which to operate
        Yields:
            eBPF program objects
        """
        vmlinux = context.modules[vmlinux_module_name]

        if not vmlinux.has_symbol("prog_idr"):
            raise exceptions.VolatilityException(
                "Cannot find the eBPF prog idr. Unsupported kernel"
            )

        prog_idr = vmlinux.object_from_symbol("prog_idr")
        for page_addr in prog_idr.get_entries():
            bpf_prog = vmlinux.object("bpf_prog", offset=page_addr, absolute=True)
            yield bpf_prog

    def _generator(self):
        for prog in self.get_ebpf_programs(self.context, self.config["kernel"]):
            prog_addr = prog.vol.offset
            prog_type = prog.get_type() or renderers.NotAvailableValue()
            prog_tag = prog.get_tag() or renderers.NotAvailableValue()
            prog_name = prog.get_name() or renderers.NotAvailableValue()
            fields = (format_hints.Hex(prog_addr), prog_name, prog_tag, prog_type)
            yield (0, fields)

    def run(self):
        headers = [
            ("Address", format_hints.Hex),
            ("Name", str),
            ("Tag", str),
            ("Type", str),
        ]
        return renderers.TreeGrid(headers, self._generator())



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import Any, List, Optional, Tuple, Union

from volatility3.cli.volshell import generic
from volatility3.framework import constants, interfaces
from volatility3.framework.configuration import requirements
from volatility3.plugins.windows import pslist


class Volshell(generic.Volshell):
    """Shell environment to directly interact with a windows memory image."""

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(name="kernel", description="Windows kernel"),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.IntRequirement(
                name="pid", description="Process ID", optional=True
            ),
            requirements.VersionRequirement(
                name="generic_volshell",
                component=generic.Volshell,
                version=(1, 0, 0),
            ),
        ] + super().get_requirements()

    def change_process(self, pid=None):
        """Change the current process and layer, based on a process ID"""
        processes = self.list_processes()
        for process in processes:
            if process.UniqueProcessId == pid:
                process_layer = process.add_process_layer()
                self.change_layer(process_layer)
                return None
        print(f"No process with process ID {pid} found")

    def list_processes(self):
        """Returns a list of EPROCESS objects from the primary layer"""
        # We always use the main kernel memory and associated symbols
        return list(
            pslist.PsList.list_processes(self.context, self.current_kernel_name)
        )

    def get_process(self, pid=None, virtaddr=None, physaddr=None):
        """Returns the _EPROCESS object that matches the pid. If a physical or a virtual address is provided, construct the _EPROCESS object at said address. Only one parameter is allowed.

        Args:
            pid (int, optional): PID / UniqueProcessId to search for.
            virtaddr (int, optional): Virtual address to construct object at
            physaddr (int, optional): Physical address to construct object at

        Returns:
            ObjectInterface: _EPROCESS Object
        """

        if sum(1 if x is not None else 0 for x in [pid, virtaddr, physaddr]) != 1:
            print("Only one parameter is accepted")
            return None

        kernel_name = self.config["kernel"]
        kernel = self.context.modules[kernel_name]

        kernel_layer_name = kernel.layer_name

        kernel_layer = self.context.layers[kernel_layer_name]
        memory_layer_name = kernel_layer.dependencies[0]

        eprocess_symbol = kernel.symbol_table_name + constants.BANG + "_EPROCESS"

        if virtaddr is not None:
            eproc = self.context.object(
                eprocess_symbol,
                layer_name=kernel_layer_name,
                offset=virtaddr,
            )

            return eproc

        if physaddr is not None:
            eproc = self.context.object(
                eprocess_symbol,
                layer_name=memory_layer_name,
                offset=physaddr,
                native_layer_name=kernel_layer_name,
            )

            return eproc

        if pid is not None:
            processes = self.list_processes()
            for process in processes:
                if process.UniqueProcessId == pid:
                    return process
            print(f"No process with process ID {pid} found")
            return None

        return None

    def construct_locals(self) -> List[Tuple[List[str], Any]]:
        result = super().construct_locals()
        result += [
            (["cp", "change_process"], self.change_process),
            (["lp", "list_processes", "ps"], self.list_processes),
            (["gp", "get_process"], self.get_process),
            (["symbols"], self.context.symbol_space[self.current_symbol_table]),
        ]
        if self.config.get("pid", None) is not None:
            self.change_process(self.config["pid"])
        return result

    def display_type(
        self,
        object: Union[
            str, interfaces.objects.ObjectInterface, interfaces.objects.Template
        ],
        offset: Optional[int] = None,
    ):
        """Display Type describes the members of a particular object in alphabetical order"""
        if isinstance(object, str):
            if constants.BANG not in object:
                object = self.current_symbol_table + constants.BANG + object
        return super().display_type(object, offset)

    def display_symbols(self, symbol_table: Optional[str] = None):
        """Prints an alphabetical list of symbols for a symbol table"""
        if symbol_table is None:
            symbol_table = self.current_symbol_table
        return super().display_symbols(symbol_table)

    @property
    def current_layer(self):
        if self.__current_layer is None:
            self.__current_layer = self.kernel.layer_name
        return self.__current_layer



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import base64
import binascii
import dataclasses
import datetime
import enum
import io
import itertools
import logging
import struct
from typing import Dict, Iterator, List, Optional, Tuple, Union

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry
from volatility3.framework.renderers import conversion
from volatility3.framework.symbols.windows.extensions import registry as reg_extensions
from volatility3.plugins import timeliner
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)

# Reference: https://cyber.wtf/2022/06/01/windows-registry-analysis-todays-episode-tasks/


class TimeMode(enum.Enum):
    """
    Enumeration containing the different time modes that a 'Time' trigger can be configured to run in.
    """

    Once = "Once"

    # run at <start_boundary> and repeat every <data1> days
    Daily = "Daily"

    # run on days of week <(data2 as day_of_week bitmap)> every <data1> weeks starting at <start_boundary>
    Weekly = "Weekly"

    # run in months <(data3 as months bitmap> on days <(data2:data1 as day in month bitmap)>
    # starting at <start_boundary>
    DaysInMonths = "Days In Months"

    # run in months <(data3 as months bitmap> in weeks <(data2 as week bitmap)>
    # on days <(data1 as day_of_week bitmap)> starting at <start_boundary>
    DaysInWeeksInMonths = "Days In Weeks in Months"

    Unknown = "Unknown"


TIME_MODE_DESCRIPTION = {
    TimeMode.Once: "Once",
    TimeMode.Daily: "Daily",
    TimeMode.Weekly: "Weekly",
    TimeMode.DaysInMonths: "Days In Months",
    TimeMode.DaysInWeeksInMonths: "Days In Weeks In Months",
    TimeMode.Unknown: "Unknown",
}


class ActionType(enum.Enum):
    """
    Enumeration that maps action types to their magic number encodings
    """

    Exe = 0x6666
    ComHandler = 0x7777
    Email = 0x8888
    MessageBox = 0x9999


class TriggerType(enum.Enum):
    """
    Enumeration that maps trigger types to their magic number encodings
    """

    WindowsNotificationFacility = 0x6666
    Session = 0x7777
    Registration = 0x8888
    Logon = 0xAAAA
    Event = 0xCCCC
    Time = 0xDDDD
    Idle = 0xEEEE
    Boot = 0xFFFF


class Weekday(enum.Enum):
    """
    Enumeration that contains bitwise values for days of the week.
    """

    Sunday = 0x1
    Monday = 0x2
    Tuesday = 0x4
    Wednesday = 0x8
    Thursday = 0x10
    Friday = 0x20
    Saturday = 0x40


class Months(enum.Enum):
    """
    Enumeration that contains bitwise values for months of the year.
    """

    January = 0x1
    February = 0x2
    March = 0x4
    April = 0x8
    May = 0x10
    June = 0x20
    July = 0x40
    August = 0x80
    September = 0x100
    October = 0x200
    November = 0x400
    December = 0x800


class SidType(enum.Enum):
    """
    Enumeration that maps SID types to their encoded integer values
    """

    User = 1
    Group = 2
    Domain = 3
    Alias = 4
    WellKnownGroup = 5
    DeletedAccount = 6
    Invalid = 7
    Unknown = 8
    Computer = 9
    Label = 10
    LogonSession = 11


@dataclasses.dataclass
class TaskSchedulerTimePeriod:
    """
    Class containing information delimiting time periods within scheduled tasks.
    """

    years: int
    months: int
    weeks: int
    days: int
    hours: int
    minutes: int
    seconds: int


JOB_BUCKET_FLAGS = {
    0x2: "Run only if idle",
    0x4: "Restart on idle",
    0x8: "Stop on idle end",
    0x10: "Disallow start if on batteries",
    0x20: "Stop if going on batteries",
    0x40: "Start when available",
    0x80: "Run only if network available",
    0x100: "Allow start on demand",
    0x200: "Wake to run",
    0x400: "Execute parallel",
    0x800: "Execute stop existing",
    0x1000: "Execute queue",
    0x2000: "Execute ignore new",
    0x4000: "Logon type s4u",
    0x10000: "Logon type InteractiveToken",
    0x40000: "Logon type Password",
    0x80000: "Logon type InteractiveTokenOrPassword",
    0x400000: "Enabled",
    0x800000: "Hidden",
    0x1000000: "Runlevel highest available",
    0x2000000: "Task",
    0x4000000: "Version",
    0x8000000: "Token SID type none",
    0x10000000: "Token SID type unrestricted",
    0x20000000: "Interval",
    0x40000000: "Allow hard terminate",
}

NULL = "\u0000"


class _ScheduledTasksReader(io.BytesIO):
    def read_task_scheduler_time(self) -> Optional[datetime.datetime]:
        _ = bool(self.read_aligned_u1())  # is_localized
        filetime = self.decode_filetime()
        if filetime is None:
            return None

        return filetime

    def read_bool(self, aligned=False) -> Optional[bool]:
        try:
            val = struct.unpack("?", self.read(1))[0]
            if aligned:
                self.seek(7)
            return val
        except struct.error:
            return None

    def decode_filetime(self) -> Optional[datetime.datetime]:
        filetime = self.read_u8()
        if filetime is None:
            return None

        if filetime == 0 or filetime == 0xFFFFFFFFFFFFFFFF:
            return None
        filetime = conversion.wintime_to_datetime(filetime)
        if isinstance(filetime, datetime.datetime):
            return filetime
        else:
            return None

    def _read_uint(
        self, size: int, format: str, aligned: bool = False
    ) -> Optional[int]:
        try:
            val = struct.unpack(format, self.read(size))[0]
            if aligned:
                self.seek(8 - size, io.SEEK_CUR)
            return val
        except struct.error:
            return None

    def read_aligned_u1(self) -> Optional[int]:
        return self._read_uint(1, "B", True)

    def read_u2(self) -> Optional[int]:
        return self._read_uint(2, "<H")

    def read_aligned_u2(self) -> Optional[int]:
        return self._read_uint(2, "<H", True)

    def read_u4(self) -> Optional[int]:
        return self._read_uint(4, "<I")

    def read_u8(self) -> Optional[int]:
        return self._read_uint(8, "<Q")

    def read_aligned_u4(self) -> Optional[int]:
        return self._read_uint(4, "<I", True)

    def read_buffer(self, aligned=False) -> Optional[bytes]:
        count = self.read_u4() if not aligned else self.read_aligned_u4()
        if count is None:
            return None
        data = self.read(count)
        if aligned:
            self.seek((8 - (count % 8)) % 8, io.SEEK_CUR)
        return data

    def read_bstring(self, aligned=False) -> Optional[str]:
        size = self.read_u4() if not aligned else self.read_aligned_u4()
        if size is None:
            return None
        try:
            raw = self.read(size)
            val = raw.decode("utf-16le", errors="replace").rstrip(NULL) or None
        except UnicodeDecodeError:
            val = None

        if aligned:
            self.seek((8 - (size % 8)) % 8, io.SEEK_CUR)

        return val

    def read_aligned_bstring_expand_sz(self) -> Optional[str]:
        sz = self.read_aligned_u4()
        if sz is None:
            return None
        byte_count = sz * 2 + 2

        if sz == 0:
            return None

        try:
            content = self.read(byte_count).decode("utf-16le")
        except UnicodeDecodeError:
            content = None

        self.seek((8 - (byte_count % 8)) % 8, io.SEEK_CUR)
        return content.rstrip("\x00") if content is not None else None

    def read_tstimeperiod(self) -> Optional[TaskSchedulerTimePeriod]:
        values = (
            self.read_u2(),
            self.read_u2(),
            self.read_u2(),
            self.read_u2(),
            self.read_u2(),
            self.read_u2(),
            self.read_u2(),
        )

        if any(value is None for value in values):
            return None

        return TaskSchedulerTimePeriod(*values)


def _build_guid_name_map(key: reg_extensions.CM_KEY_NODE) -> Dict[str, str]:
    mapping = {}
    task_id_value = None
    for value in key.get_values():
        try:
            if value.get_name() == "Id":
                task_id_value = value
                break
        except (
            exceptions.InvalidAddressException,
            registry.RegistryException,
        ):
            continue

    if (
        task_id_value is not None
        and task_id_value.get_type() == reg_extensions.RegValueTypes.REG_SZ
    ):
        try:
            id_str = task_id_value.decode_data()
        except exceptions.InvalidAddressException:
            id_str = None

        try:
            if isinstance(id_str, bytes):
                mapping[id_str.decode("utf-16le", errors="replace").rstrip(NULL)] = str(
                    key.get_name()
                )
        except (
            exceptions.InvalidAddressException,
            registry.RegistryException,
        ) as excp:
            vollog.debug(f"Exception occurred while decoding id_str: {excp}")

    for subkey in key.get_subkeys():
        mapping.update(_build_guid_name_map(subkey))
    return mapping


@dataclasses.dataclass
class TaskAction:
    action_type: ActionType
    action: str
    action_args: Optional[str]
    working_directory: Optional[str]

    @classmethod
    def decode_messagebox_action(
        cls, reader: _ScheduledTasksReader
    ) -> Optional["TaskAction"]:
        caption, content = reader.read_bstring(), reader.read_bstring()
        return cls(
            ActionType.MessageBox,
            f'"{caption or "<Unknown>"}": {content or "<Unknown>"}',
            None,
            None,
        )

    @classmethod
    def _decode_exe_action(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskAction"]:
        command = reader.read_bstring()
        args = reader.read_bstring()
        if command is None or args is None:
            return None

        workdir = reader.read_bstring()
        if version == 3:
            _flags = reader.read_u2()

        return cls(ActionType.Exe, command, args, workdir)

    @classmethod
    def _decode_email_action(
        cls, reader: _ScheduledTasksReader
    ) -> Optional["TaskAction"]:
        props = {
            "From": reader.read_bstring(),
            "To": reader.read_bstring(),
            "Cc": reader.read_bstring(),
            "Bcc": reader.read_bstring(),
            "Reply_to": reader.read_bstring(),
            "Server": reader.read_bstring(),
            "Subject": reader.read_bstring(),
            "Body": reader.read_bstring(),
        }

        num_attachment_filenames = reader.read_u4()
        if num_attachment_filenames is not None:
            attachment_filenames = [
                reader.read_bstring() for _ in range(num_attachment_filenames)
            ]

            props["Attachments"] = (
                "<"
                + ", ".join(
                    filename
                    for filename in attachment_filenames
                    if filename is not None
                )
                + ">"
            )

        num_headers = reader.read_u4()
        if num_headers is not None:
            headers = [
                (reader.read_bstring(), reader.read_bstring())
                for _ in range(num_headers)
            ]

            props["Headers"] = (
                "<"
                + ", ".join(
                    f"{field}: {value}"
                    for field, value in headers
                    if field is not None and value is not None
                )
                + ">"
            )

        cls(
            ActionType.Email,
            ", ".join(
                f"{key}: {value}" for key, value in props.items() if value is not None
            ),
            None,
            None,
        )

    @classmethod
    def _decode_comhandler_action(
        cls, reader: _ScheduledTasksReader
    ) -> Optional["TaskAction"]:
        guid_raw = reader.read(16)
        if not guid_raw and len(guid_raw) == 16:
            return None
        clsid = conversion.windows_bytes_to_guid(guid_raw)
        args = reader.read_bstring()

        return cls(ActionType.ComHandler, clsid, args, None)


@dataclasses.dataclass
class _ScheduledTaskEntry:
    name: Union[str, interfaces.renderers.BaseAbsentValue]
    principal_id: Union[str, interfaces.renderers.BaseAbsentValue]
    display_name: Union[str, interfaces.renderers.BaseAbsentValue]
    enabled: Union[bool, interfaces.renderers.BaseAbsentValue]
    creation_time: Union[datetime.datetime, interfaces.renderers.BaseAbsentValue]
    last_run_time: Union[datetime.datetime, interfaces.renderers.BaseAbsentValue]
    last_successful_run_time: Union[
        datetime.datetime, interfaces.renderers.BaseAbsentValue
    ]
    trigger_type: Union[str, interfaces.renderers.BaseAbsentValue]
    trigger_description: Union[str, interfaces.renderers.BaseAbsentValue]
    action_type: Union[str, interfaces.renderers.BaseAbsentValue]
    action_description: Union[str, interfaces.renderers.BaseAbsentValue]
    action_args: Union[str, interfaces.renderers.BaseAbsentValue]
    action_context: Union[str, interfaces.renderers.BaseAbsentValue]
    working_directory: Union[str, interfaces.renderers.BaseAbsentValue]
    guid: str


@dataclasses.dataclass
class _JobSchedule:
    start_boundary: Optional[datetime.datetime]
    end_boundary: Optional[datetime.datetime]
    repetition_interval_secs: Optional[int]
    repetition_duration_secs: Optional[int]
    execution_time_limit_secs: Optional[int]
    mode: Optional[TimeMode]
    data1: Optional[int]
    data2: Optional[int]
    data3: Optional[int]
    stop_tasks_at_duration_end: Optional[int]
    is_enabled: Optional[bool]
    max_delay_seconds: Optional[int]

    def get_description(self) -> Optional[str]:
        if self.mode == TimeMode.Once:
            return "Run one time starting at {}".format(
                self.start_boundary.isoformat()
                if self.start_boundary is not None
                else "<UNKNOWN>"
            )

        elif self.mode == TimeMode.Daily:
            if self.data1 is None:
                return None
            return "Run at {} and repeat every {} days".format(
                (
                    self.start_boundary.isoformat()
                    if self.start_boundary is not None
                    else "<UNKNOWN>"
                ),
                self.data1,
            )

        elif self.mode == TimeMode.Weekly:
            if self.data2 is None:
                return None

            days = [k.name for k in Weekday if k.value & self.data2]
            return "Run on {} every {} weeks starting at {}".format(
                ", ".join(days),
                self.data1,
                (
                    self.start_boundary.isoformat()
                    if self.start_boundary is not None
                    else "<UNKNOWN>"
                ),
            )
        elif self.mode == TimeMode.DaysInMonths:
            if self.data2 is None or self.data1 is None or self.data3 is None:
                return None
            months = [month.name for month in Months if month.value & self.data3]
            days_bitmap = (self.data2 << 16) + self.data1
            days = [str(v + 1) for v in range(31) if (1 << v) & days_bitmap]
            return "Run in months {} on days {} starting at {}".format(
                ", ".join(months),
                ", ".join(days),
                (
                    self.start_boundary.isoformat()
                    if self.start_boundary is not None
                    else "<UNKNOWN>"
                ),
            )
        elif self.mode == TimeMode.DaysInWeeksInMonths:
            if self.data1 is None or self.data2 is None or self.data3 is None:
                return None

            months = [month.name for month in Months if month.value & self.data3]
            weeks = [str(v + 1) for v in range(5) if (v << 1) & self.data2]
            days = [day.name for day in Weekday if day.value & self.data1]
            return "Run in months {} in weeks {} on days {} starting at {}".format(
                ", ".join(months),
                ", ".join(weeks),
                ", ".join(days),
                (
                    self.start_boundary.isoformat()
                    if self.start_boundary is not None
                    else "<UNKNOWN>"
                ),
            )
        else:
            return None

    @classmethod
    def decode(cls, reader: _ScheduledTasksReader) -> Optional["_JobSchedule"]:
        start_boundary = reader.read_task_scheduler_time()
        end_boundary = reader.read_task_scheduler_time()

        _ = reader.read_task_scheduler_time()
        repetition_interval_secs = reader.read_u4()
        repetition_duration_secs = reader.read_u4()
        execution_time_limit_secs = reader.read_u4()
        mode_index = reader.read_u4()
        if mode_index is not None:
            try:
                mode = TimeMode(mode_index)
            except ValueError:
                mode = TimeMode.Unknown
        else:
            mode = None

        data1 = reader.read_u2()
        data2 = reader.read_u2()
        data3 = reader.read_u2()

        reader.seek(2, io.SEEK_CUR)  # pad
        stop_tasks_at_duration_end = reader.read_bool()
        is_enabled = reader.read_bool()
        reader.seek(6, io.SEEK_CUR)  # pad (2) + unknown (4)
        max_delay_seconds = reader.read_u4()
        reader.seek(4, io.SEEK_CUR)  # pad

        return cls(
            start_boundary,
            end_boundary,
            repetition_interval_secs,
            repetition_duration_secs,
            execution_time_limit_secs,
            mode,
            data1,
            data2,
            data3,
            stop_tasks_at_duration_end,
            is_enabled,
            max_delay_seconds,
        )


def decode_sid(data: bytes) -> Optional[str]:
    """
    Decodes a windows SID from variable-length raw bytes

    Returns the string representation of the SID if decoding was successful, or None
    if the data could not be parsed due to an insufficient number of bytes.
    """
    try:
        revision, subid_count, id_authority = struct.unpack(
            ">BBQ", data[:2] + b"\x00\x00" + data[2:8]
        )
        subauthorities = struct.unpack(
            "<" + "I" * subid_count, data[8 : 8 + subid_count * 4]
        )
        sid_string = "S-" + "-".join(
            [str(item) for item in [revision, id_authority] + list(subauthorities)]
        )
    except struct.error:
        return None

    return sid_string


@dataclasses.dataclass
class UserInfo:
    sid_type: Optional[SidType]
    sid: Optional[str]
    username: Optional[str]

    @classmethod
    def _decode(cls, reader: _ScheduledTasksReader) -> Optional["UserInfo"]:
        skip_user = reader.read_aligned_u1() != 0
        if not skip_user:
            skip_sid = reader.read_aligned_u1() != 0
        else:
            skip_sid = None

        sid_type = None
        sid = None
        if not skip_user and not skip_sid:
            try:
                sid_type = SidType(reader.read_aligned_u4())
            except ValueError:
                sid_type = SidType.Unknown

            sid_raw = reader.read_buffer(aligned=True)
            if sid_raw is None:
                return None
            sid = decode_sid(sid_raw)

        username = reader.read_bstring(aligned=True) if not skip_user else None

        return UserInfo(sid_type, sid, username)


@dataclasses.dataclass
class OptionalSettings:
    IdleDurationSeconds: int
    idleWaitTimeoutSeconds: int
    ExecutionTimeLimitSeconds: int
    DeleteExpiredTaskAfter: int
    Priority: int
    RestartOnFailureDelay: int
    RestartOnFailureRetries: int
    NetworkId: bytes
    Privileges: Optional[List[str]]
    Periodicity: Optional[TaskSchedulerTimePeriod]
    Deadline: Optional[TaskSchedulerTimePeriod]
    Exclusive: Optional[bool]

    @classmethod
    def _decode(cls, reader: _ScheduledTasksReader) -> Optional["OptionalSettings"]:
        LEN_WITH_PRIVILEGES = 0x38
        LEN_WITH_TIME_PERIODS = 0x58
        length = reader.read_aligned_u4()
        if length == 0:
            return None

        base_values = (
            reader.read_u4(),
            reader.read_u4(),
            reader.read_u4(),
            reader.read_u4(),
            reader.read_u4(),
            reader.read_u4(),
            reader.read_u4(),
            binascii.hexlify(reader.read(16)),
        )

        if any(value is None for value in base_values):
            return None

        reader.seek(4, io.SEEK_CUR)  # padding

        privileges = None
        periodicity = None
        deadline = None
        exclusive = None
        if length == LEN_WITH_PRIVILEGES or length == LEN_WITH_TIME_PERIODS:
            privileges_raw = reader.read_u8()
            if privileges_raw is None:
                return None
            privileges = [
                priv.name for priv in Privileges if priv.value & privileges_raw
            ]
        if length == LEN_WITH_TIME_PERIODS:
            periodicity = reader.read_tstimeperiod()
            deadline = reader.read_tstimeperiod()
            exclusive = reader.read_bool()
            reader.seek(3, io.SEEK_CUR)  # padding

        return OptionalSettings(
            *base_values, privileges, periodicity, deadline, exclusive
        )


@dataclasses.dataclass
class JobBucket:
    flags: List[str]
    crc32: int
    principal_id: Optional[str]
    display_name: Optional[str]
    user_info: Optional[UserInfo]
    optional_settings: Optional[OptionalSettings]

    @classmethod
    def _decode(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["JobBucket"]:
        flags_raw = reader.read_aligned_u4()
        if flags_raw is None:
            return None
        flags = [y for x, y in JOB_BUCKET_FLAGS.items() if x & flags_raw]
        crc32 = reader.read_aligned_u4()
        if crc32 is None:
            return None

        principal_id = None
        display_name = None
        if version >= 0x16:
            principal_id = reader.read_bstring(aligned=True)
        if version >= 0x17:
            display_name = reader.read_bstring(aligned=True)

        user_info = UserInfo._decode(reader)
        optional_settings = OptionalSettings._decode(reader)

        return JobBucket(
            flags, crc32, principal_id, display_name, user_info, optional_settings
        )


class Privileges(enum.Enum):
    SeCreateTokenPrivilege = 0x4
    SeAssignPrimaryTokenPrivilege = 0x8
    SeLockMemoryPrivilege = 0x10
    SeIncreaseQuotaPrivilege = 0x20
    SeMachineAccountPrivilege = 0x40
    SeTcbPrivilege = 0x80
    SeSecurityPrivilege = 0x100
    SeTakeOwnershipPrivilege = 0x200
    SeLoadDriverPrivilege = 0x400
    SeSystemProfilePrivilege = 0x800
    SeSystemtimePrivilege = 0x1000
    SeProfileSingleProcessPrivilege = 0x2000
    SeIncreaseBasePriorityPrivilege = 0x4000
    SeCreatePagefilePrivilege = 0x8000
    SeCreatePermanentPrivilege = 0x10000
    SeBackupPrivilege = 0x20000
    SeRestorePrivilege = 0x40000
    SeShutdownPrivilege = 0x80000
    SeDebugPrivilege = 0x100000
    SeAuditPrivilege = 0x200000
    SeSystemEnvironmentPrivilege = 0x400000
    SeChangeNotifyPrivilege = 0x800000
    SeRemoteShutdownPrivilege = 0x1000000
    SeUndockPrivilege = 0x2000000
    SeSyncAgentPrivilege = 0x4000000
    SeEnableDelegationPrivilege = 0x8000000
    SeManageVolumePrivilege = 0x10000000
    SeImpersonatePrivilege = 0x20000000
    SeCreateGlobalPrivilege = 0x40000000
    SeTrustedCredManAccessPrivilege = 0x80000000
    SeRelabelPrivilege = 0x100000000
    SeIncreaseWorkingSetPrivilege = 0x200000000
    SeTimeZonePrivilege = 0x400000000
    SeCreateSymbolicLinkPrivilege = 0x800000000
    SeDelegateSessionUserImpersonatePrivilege = 0x1000000000


class SessionState(enum.Enum):
    ConsoleConnect = 1
    ConsoleDisconnect = 2
    RemoteConnect = 3
    RemoteDisconnect = 4
    SessionLock = 5
    SessionUnlock = 6
    Unknown = "Unknown"


@dataclasses.dataclass
class TaskTrigger:
    start_boundary: Optional[datetime.datetime]
    end_boundary: Optional[datetime.datetime]
    repetition_interval_seconds: Optional[int]
    enabled: Optional[bool]
    trigger_type: TriggerType
    description: Optional[str]

    @classmethod
    def _decode_generic_trigger(
        cls, reader: _ScheduledTasksReader, version: int, trigger_type: TriggerType
    ) -> Optional["TaskTrigger"]:
        start_boundary = reader.read_task_scheduler_time()
        end_boundary = reader.read_task_scheduler_time()

        _ = reader.read_u4()  # delay seconds
        _ = reader.read_u4()  # timeout seconds

        repetition_interval_secs = reader.read_u4()
        _ = reader.read_u4()  # repetition duration seconds
        _ = reader.read_u4()  # repetition duration seconds 2

        _ = reader.read_bool()  # stop at duration end
        reader.seek(3, io.SEEK_CUR)
        trigger_enabled = bool(reader.read_aligned_u1())
        reader.seek(8, io.SEEK_CUR)  # unknown field

        if version >= 0x16:
            cur = reader.tell()
            _ = reader.read_bstring()  # trigger id
            reader.seek((8 - (reader.tell() - cur)) % 8, io.SEEK_CUR)  # pad to block

        return cls(
            start_boundary,
            end_boundary,
            repetition_interval_secs,
            trigger_enabled,
            trigger_type,
            f"{trigger_type.name} trigger",
        )

    @classmethod
    def _decode_logon_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        base = cls._decode_generic_trigger(reader, version, TriggerType.Logon)
        if base is None:
            return None

        user = UserInfo._decode(reader)
        if user is not None and user.username is not None:
            base.description = f"{user.username}: {user.sid} ({user.sid_type})"

        return base

    @classmethod
    def _decode_session_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        base = cls._decode_generic_trigger(reader, version, TriggerType.Session)
        if base is None:
            return None
        session_type_raw = reader.read_u4()
        reader.seek(4, io.SEEK_CUR)

        try:
            session_type = SessionState(session_type_raw)
        except ValueError:
            session_type = SessionState.Unknown

        user_info = UserInfo._decode(reader)
        if user_info is not None and user_info.username is not None:
            base.description = f"{session_type.name} for user {user_info.username}"
        else:
            base.description = session_type.name

        return base

    @classmethod
    def _decode_time_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        job_schedule = _JobSchedule.decode(reader)
        if job_schedule is None:
            return None

        if version >= 0x16:
            cur = reader.tell()
            _ = reader.read_bstring()  # trigger id
            reader.seek((8 - (reader.tell() - cur)) % 8, io.SEEK_CUR)  # pad to block

        return cls(
            job_schedule.start_boundary,
            job_schedule.end_boundary,
            job_schedule.repetition_interval_secs,
            job_schedule.is_enabled,
            TriggerType.Time,
            job_schedule.get_description() or None,
        )

    @classmethod
    def _decode_event_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        base = cls._decode_generic_trigger(reader, version, TriggerType.Event)
        if base is None:
            return base

        subscription = reader.read_aligned_bstring_expand_sz()
        reader.seek(8, io.SEEK_CUR)  # 2 4-byte unknown fields
        reader.read_aligned_bstring_expand_sz()  # another unknown field
        len_value_queries = reader.read_aligned_u4()

        if len_value_queries is None:
            return base

        queries = [
            (
                reader.read_aligned_bstring_expand_sz(),
                reader.read_aligned_bstring_expand_sz(),
            )
            for _ in range(len_value_queries)
        ]
        valid = [(k, v) for (k, v) in queries if k is not None and v is not None]
        if base.description is None:
            base.description = "Event Trigger"
        base.description += f": Subscription: {subscription}, Queries: {str(valid)}"
        return base

    @classmethod
    def _decode_boot_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        return cls._decode_generic_trigger(reader, version, TriggerType.Boot)

    @classmethod
    def _decode_wnf_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        base = cls._decode_generic_trigger(
            reader, version, TriggerType.WindowsNotificationFacility
        )
        if base is None:
            return None

        state_name = binascii.hexlify(reader.read(8)).decode("ascii")
        datalen = reader.read_aligned_u4()
        _ = base64.b64encode(reader.read(datalen))  # state binary data
        base.description = f"WNF state {state_name}"
        return base

    @classmethod
    def _decode_idle_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        return cls._decode_generic_trigger(reader, version, TriggerType.Logon)

    @classmethod
    def _decode_registration_trigger(
        cls, reader: _ScheduledTasksReader, version: int
    ) -> Optional["TaskTrigger"]:
        return cls._decode_generic_trigger(reader, version, TriggerType.Logon)


@dataclasses.dataclass
class TriggerSet:
    job_bucket: JobBucket
    triggers: List[TaskTrigger]

    @classmethod
    def decode(cls, data) -> Optional["TriggerSet"]:
        reader = _ScheduledTasksReader(data)

        version = reader.read_aligned_u1()
        _ = reader.read_task_scheduler_time()  # start boundary
        _ = reader.read_task_scheduler_time()  # end_boundary

        if version is None:
            return None

        job_bucket = JobBucket._decode(reader, version)
        if job_bucket is None:
            return None

        triggers = []

        while True:
            magic = reader.read_aligned_u4()
            if magic is None:
                break
            try:
                trigger_type = TriggerType(magic)
            except ValueError:
                vollog.warning(f"Invalid trigger magic {hex(magic)}")
                break

            if trigger_type == TriggerType.Logon:
                trigger = TaskTrigger._decode_logon_trigger(reader, version)
            elif trigger_type == TriggerType.Session:
                trigger = TaskTrigger._decode_session_trigger(reader, version)
            elif trigger_type == TriggerType.WindowsNotificationFacility:
                trigger = TaskTrigger._decode_wnf_trigger(reader, version)
            elif trigger_type == TriggerType.Boot:
                trigger = TaskTrigger._decode_boot_trigger(reader, version)
            elif trigger_type == TriggerType.Registration:
                trigger = TaskTrigger._decode_registration_trigger(reader, version)
            elif trigger_type == TriggerType.Event:
                trigger = TaskTrigger._decode_event_trigger(reader, version)
            elif trigger_type == TriggerType.Idle:
                trigger = TaskTrigger._decode_idle_trigger(reader, version)
            elif trigger_type == TriggerType.Time:
                trigger = TaskTrigger._decode_time_trigger(reader, version)
            else:
                vollog.warning(
                    f"Invalid trigger magic {hex(magic)} encountered at offset {hex(reader.tell() - 8)}, stopping parsing"
                )
                break
            triggers.append(trigger)

        return cls(job_bucket, triggers)


@dataclasses.dataclass
class ActionSet:
    actions: List[TaskAction]
    context: Optional[str]

    @classmethod
    def decode(cls, data: bytes) -> Optional["ActionSet"]:
        reader = _ScheduledTasksReader(data)
        actions = []

        version = reader.read_u2()
        if version is None:
            return None

        if version in [2, 3]:
            action_context = reader.read_bstring()
        else:
            action_context = None

        while True:
            magic = reader.read_u2()
            if magic is None:
                break

            _ = (
                reader.read_bstring()
            )  # action identifier, usually (but not always) empty

            if magic == ActionType.Email.value:
                action = TaskAction._decode_email_action(reader)
            elif magic == ActionType.Exe.value:
                action = TaskAction._decode_exe_action(reader, version)
            elif magic == ActionType.ComHandler.value:
                action = TaskAction._decode_comhandler_action(reader)
            elif magic == ActionType.MessageBox.value:
                action = TaskAction.decode_messagebox_action(reader)
            else:
                break
            actions.append(action)

        return cls(actions, action_context)


@dataclasses.dataclass
class DynamicInfo:
    """
    Contains information about execution history for this task,
    including timestamps and the last error code
    """

    creation_time: Optional[datetime.datetime]
    last_run_time: Optional[datetime.datetime]
    last_successful_run_time: Optional[datetime.datetime]
    last_error_code: Optional[int]

    @classmethod
    def decode(cls, data: bytes) -> Optional["DynamicInfo"]:
        """
        Decodes a DynamicInfo structure from RegBin value data.
        Raises a `ScheduledTaskDecodingError` if the magic bytes are invalid, but otherwise
        attempts to decode as much as possible without returning an error.
        """
        DYNAMICINFO_MAGIC = 3

        reader = _ScheduledTasksReader(data)
        magic = reader.read_u4()
        if magic != DYNAMICINFO_MAGIC:
            return None

        creation_time = reader.decode_filetime()
        last_run_time = reader.decode_filetime()

        reader.seek(4, io.SEEK_CUR)  # deprecated field 'TaskState'

        last_error_code = reader.read_u4()
        last_success_time = reader.decode_filetime()

        vollog.debug((creation_time, last_run_time, last_success_time))

        return cls(
            last_run_time,
            creation_time,
            last_success_time,
            last_error_code,
        )


class ScheduledTasks(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Decodes scheduled task information from the Windows registry, including
    information about triggers, actions, run times, and creation times."""

    _required_framework_version = (2, 11, 0)
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel33", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
        ]

    def generate_timeline(
        self,
    ) -> Iterator[Tuple[str, timeliner.TimeLinerType, datetime.datetime]]:
        for _, task in self._generator():
            if isinstance(task.last_run_time, datetime.datetime):
                yield (
                    f"ScheduledTasks: task action {task.action_description} with trigger {task.trigger_description} ran",
                    timeliner.TimeLinerType.ACCESSED,
                    task.last_run_time,
                )
            if isinstance(task.last_successful_run_time, datetime.datetime):
                yield (
                    f"ScheduledTasks: task action {task.action_description} with trigger {task.trigger_description} ran successfully",
                    timeliner.TimeLinerType.ACCESSED,
                    task.last_successful_run_time,
                )
            if isinstance(task.creation_time, datetime.datetime):
                yield (
                    f"ScheduledTasks: Creation Time for task {task.guid} with trigger {task.trigger_description or '<UNKNOWN>'}",
                    timeliner.TimeLinerType.CREATED,
                    task.creation_time,
                )

    @classmethod
    def get_software_hive(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
    ) -> Optional[registry.RegistryHive]:
        """Retrieves the `Amcache.hve` registry hive from the kernel module, if it can be located."""
        return next(
            hivelist.HiveList.list_hives(
                context=context,
                base_config_path=interfaces.configuration.path_join(
                    config_path, "hivelist"
                ),
                kernel_module_name=kernel_module_name,
                filter_string="SOFTWARE",
            ),
            None,
        )

    @classmethod
    def parse_actions_value(
        cls, actions_value: reg_extensions.CM_KEY_VALUE
    ) -> Optional[ActionSet]:
        """Parses File entries from the Windows 8 `Root\\File` key.

        :param programs_key: The `Root\\File` registry key.

        :return: An iterator of tuples, where the first member is the program ID string for
        correlating `Root\\Program` entries, and the second member is the `AmcacheEntry`.
        """
        try:
            data = actions_value.decode_data()
        except exceptions.InvalidAddressException:
            data = None

        if not isinstance(data, bytes):
            return None

        return ActionSet.decode(data)

    @classmethod
    def parse_triggers_value(
        cls, triggers_value: reg_extensions.CM_KEY_VALUE
    ) -> Optional[TriggerSet]:
        try:
            data = triggers_value.decode_data()
        except exceptions.InvalidAddressException:
            data = None

        if not isinstance(data, bytes):
            return None

        return TriggerSet.decode(data)

    @classmethod
    def parse_dynamic_info_value(
        cls, dyn_info_value: reg_extensions.CM_KEY_VALUE
    ) -> Optional[DynamicInfo]:
        try:
            data = dyn_info_value.decode_data()
        except exceptions.InvalidAddressException:
            data = None

        if not isinstance(data, bytes):
            return None

        return DynamicInfo.decode(data)

    @classmethod
    def _get_task_keys(
        cls, software_hive: registry.RegistryHive
    ) -> Tuple[
        Optional[reg_extensions.CM_KEY_NODE], Optional[reg_extensions.CM_KEY_NODE]
    ]:
        try:
            task_key = software_hive.get_key(
                "Microsoft\\Windows NT\\CurrentVersion\\Schedule\\TaskCache\\Tasks"
            )
        except (KeyError, registry.RegistryException):
            task_key = None

        try:
            task_tree = software_hive.get_key(
                "Microsoft\\Windows NT\\CurrentVersion\\Schedule\\TaskCache\\Tree"
            )
        except (KeyError, registry.RegistryException):
            task_tree = None

        return (task_key, task_tree)  # type: ignore

    @classmethod
    def _parse_task_key(
        cls, key: reg_extensions.CM_KEY_NODE, guid_mapping: Dict[str, str]
    ) -> Iterator[_ScheduledTaskEntry]:
        values = {}
        for value in key.get_values():
            try:
                name = str(value.get_name())
            except (
                exceptions.InvalidAddressException,
                registry.RegistryException,
            ):
                continue

            if name in ["Actions", "Triggers", "DynamicInfo"]:
                values[name] = value

        try:
            key_name = str(key.get_name())
        except (
            exceptions.InvalidAddressException,
            registry.RegistryException,
        ):
            key_name = None

        try:
            task_name = guid_mapping.get(key_name, renderers.NotAvailableValue())
        except (
            exceptions.InvalidAddressException,
            registry.RegistryException,
        ):
            task_name = renderers.NotAvailableValue()

        try:
            action_set = cls.parse_actions_value(values["Actions"])
        except KeyError:
            vollog.debug("Failed to get Actions value")
            action_set = None

        try:
            triggers_value = values["Triggers"]
            trigger_set = cls.parse_triggers_value(triggers_value)
        except KeyError:
            vollog.debug("Failed to get Triggers value")
            trigger_set = None

        if trigger_set is not None:
            vollog.debug("Parsed triggers successfully")

            principal_id = (
                trigger_set.job_bucket.principal_id or renderers.NotAvailableValue()
            )
            display_name = (
                trigger_set.job_bucket.display_name or renderers.NotAvailableValue()
            )
        else:
            vollog.debug("Failed to parse triggers")

            principal_id = renderers.NotAvailableValue()
            display_name = renderers.NotAvailableValue()

        try:
            dynamic_info = cls.parse_dynamic_info_value(values["DynamicInfo"])
        except KeyError:
            vollog.debug("DynamicInfo value not found")
            dynamic_info = None

        vollog.debug(dynamic_info)

        creation_time = dynamic_info.creation_time if dynamic_info is not None else None
        last_run_time = dynamic_info.last_run_time if dynamic_info is not None else None
        last_successful_run_time = (
            dynamic_info.last_successful_run_time if dynamic_info is not None else None
        )

        all_triggers = (
            trigger_set.triggers or [None] if trigger_set is not None else [None]
        )

        all_actions = action_set.actions or [None] if action_set is not None else [None]

        for action, trigger in itertools.product(all_actions, all_triggers):
            if action is not None:
                if action.action_type in (
                    ActionType.Exe,
                    ActionType.ComHandler,
                ):
                    if action.action_args is None:
                        args = renderers.NotAvailableValue()
                    else:
                        args = action.action_args
                else:
                    args = renderers.NotApplicableValue()

                if action.action_type == ActionType.Exe:
                    working_directory = (
                        action.working_directory or renderers.NotAvailableValue()
                    )
                else:
                    working_directory = renderers.NotApplicableValue()

            else:
                args = renderers.NotAvailableValue()
                working_directory = renderers.NotAvailableValue()

            if trigger is not None and trigger.enabled is not None:
                enabled = trigger.enabled
            else:
                enabled = renderers.NotAvailableValue()

            yield _ScheduledTaskEntry(
                task_name,
                principal_id,
                display_name,
                enabled,
                creation_time or renderers.NotAvailableValue(),
                last_run_time or renderers.NotAvailableValue(),
                last_successful_run_time or renderers.NotAvailableValue(),
                (
                    trigger.trigger_type.name
                    if trigger is not None
                    else renderers.NotAvailableValue()
                ),
                (
                    trigger.description or renderers.NotAvailableValue()
                    if trigger is not None
                    else renderers.NotAvailableValue()
                ),
                (
                    action.action_type.name
                    if action is not None
                    else renderers.NotAvailableValue()
                ),
                (
                    action.action
                    if action is not None
                    else renderers.NotAvailableValue()
                ),
                args,
                (
                    action_set.context
                    if (action_set is not None and action_set.context is not None)
                    else renderers.NotAvailableValue()
                ),
                working_directory,
                key_name or renderers.NotAvailableValue(),
            )

    def _generator(self) -> Iterator[Tuple[int, _ScheduledTaskEntry]]:
        # Building the dictionary ahead of time is much better for performance
        # vs looking up each service's DLL individually.
        software_hive = self.get_software_hive(
            self.context, self.config_path, self.config["kernel"]
        )
        if software_hive is None:
            vollog.warning("Failed to get SOFTWARE hive")
            return

        task_key_root, task_tree = self._get_task_keys(software_hive)
        if task_key_root is None:
            vollog.warning("Failed to get 'Tasks' key")
            return

        if task_tree is not None:
            task_name_map = _build_guid_name_map(task_tree)
        else:
            vollog.info("'Tree' key not found, can't map GUIDs to task names")
            task_name_map = {}

        for key in task_key_root.get_subkeys():
            for task in self._parse_task_key(key, task_name_map):
                yield 0, task

    def run(self):
        return renderers.TreeGrid(
            [
                ("Task Name", str),
                ("Principal ID", str),
                ("Display Name", str),
                ("Enabled", bool),
                ("Creation Time", datetime.datetime),
                ("Last Run Time", datetime.datetime),
                ("Last Successful Run Time", datetime.datetime),
                ("Trigger Type", str),
                ("Trigger Description", str),
                ("Action Type", str),
                ("Action", str),
                ("Action Arguments", str),
                ("Action Context", str),
                ("Working Directory", str),
                ("Key Name", str),
            ],
            (
                (indent, dataclasses.astuple(entry))
                for indent, entry in self._generator()
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module containing a plugin that recovers bash command history
from bash process memory."""

import datetime
import struct
from typing import List

from volatility3.framework import constants, renderers, symbols, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.layers import scanners
from volatility3.framework.objects import utility
from volatility3.framework.symbols.linux import bash
from volatility3.plugins import timeliner
from volatility3.plugins.linux import pslist


class Bash(plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Recovers bash command history from memory."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 2)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="multi_string_scanner",
                component=scanners.MultiStringScanner,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
        ]

    def _generator(self, tasks):
        vmlinux = self.context.modules[self.config["kernel"]]
        is_32bit = not symbols.symbol_table_is_64bit(
            context=self.context, symbol_table_name=vmlinux.symbol_table_name
        )
        if is_32bit:
            pack_format = "I"
            bash_json_file = "bash32"
        else:
            pack_format = "Q"
            bash_json_file = "bash64"

        bash_table_name = bash.BashIntermedSymbols.create(
            self.context, self.config_path, "linux", bash_json_file
        )

        ts_offset = self.context.symbol_space.get_type(
            bash_table_name + constants.BANG + "hist_entry"
        ).relative_child_offset("timestamp")

        for task in tasks:
            task_name = utility.array_to_string(task.comm)
            if task_name not in ["bash", "sh", "dash"]:
                continue

            proc_layer_name = task.add_process_layer()
            if not proc_layer_name:
                continue

            proc_layer = self.context.layers[proc_layer_name]

            bang_addrs = []

            # get task memory sections to be used by scanners
            task_memory_sections = [
                section for section in task.get_process_memory_sections(heap_only=True)
            ]

            # find '#' values on the heap
            for address in proc_layer.scan(
                self.context,
                scanners.BytesScanner(b"#"),
                sections=task_memory_sections,
            ):
                bang_addrs.append(struct.pack(pack_format, address))

            history_entries = []

            if bang_addrs:
                for address, _ in proc_layer.scan(
                    self.context,
                    scanners.MultiStringScanner(bang_addrs),
                    sections=task_memory_sections,
                ):
                    hist = self.context.object(
                        bash_table_name + constants.BANG + "hist_entry",
                        offset=address - ts_offset,
                        layer_name=proc_layer_name,
                    )

                    if hist.is_valid():
                        history_entries.append(hist)

            for hist in sorted(history_entries, key=lambda x: x.get_time_as_integer()):
                yield (
                    0,
                    (task.pid, task_name, hist.get_time_object(), hist.get_command()),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("CommandTime", datetime.datetime),
                ("Command", str),
            ],
            self._generator(
                pslist.PsList.list_tasks(
                    self.context, self.config["kernel"], filter_func=filter_func
                )
            ),
        )

    def generate_timeline(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        for row in self._generator(
            pslist.PsList.list_tasks(
                self.context, self.config["kernel"], filter_func=filter_func
            )
        ):
            _depth, row_data = row
            description = f'{row_data[0]} ({row_data[1]}): "{row_data[3]}"'
            yield (description, timeliner.TimeLinerType.CREATED, row_data[2])



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""In-memory artifacts from OSX systems."""

from typing import Iterator, Tuple, Any, Generator, List

from volatility3.framework import exceptions, renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.plugins.mac import pslist


class Psaux(plugins.PluginInterface):
    """Recovers program command line arguments."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    def _generator(
        self, tasks: Iterator[Any]
    ) -> Generator[Tuple[int, Tuple[int, str, int, str]], None, None]:
        for task in tasks:
            proc_layer_name = task.add_process_layer()
            if proc_layer_name is None:
                continue

            proc_layer = self.context.layers[proc_layer_name]

            argsstart = task.user_stack - task.p_argslen

            if (
                not proc_layer.is_valid(argsstart)
                or task.p_argslen == 0
                or task.p_argc == 0
            ):
                continue

            # Add one because the first two are usually duplicates
            argc = task.p_argc + 1

            # smear protection
            if argc > 1024:
                continue

            task_name = utility.array_to_string(task.p_comm)

            args: List[bytes] = []

            while argc > 0:
                try:
                    arg = proc_layer.read(argsstart, 256)
                except exceptions.InvalidAddressException:
                    break

                idx = arg.find(b"\x00")
                if idx != -1:
                    arg = arg[:idx]

                argsstart += len(str(arg)) + 1

                # deal with the stupid alignment (leading nulls) and arg duplication
                if len(args) == 0:
                    while argsstart < task.user_stack:
                        try:
                            check = proc_layer.read(argsstart, 1)
                        except exceptions.InvalidAddressException:
                            break

                        if check != b"\x00":
                            break

                        argsstart = argsstart + 1

                    args.append(arg)

                # also check for initial duplicates since OS X is painful
                elif arg != args[0]:
                    args.append(arg)

                argc = argc - 1

            args_str = " ".join([s.decode("utf-8", errors="replace") for s in args])

            yield (0, (task.p_pid, task_name, task.p_argc, args_str))

    def run(self) -> renderers.TreeGrid:
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        list_tasks = pslist.PsList.get_list_tasks(
            self.config.get("pslist_method", pslist.PsList.pslist_methods[0])
        )

        return renderers.TreeGrid(
            [("PID", int), ("Process", str), ("Argc", int), ("Arguments", str)],
            self._generator(
                list_tasks(self.context, self.config["kernel"], filter_func=filter_func)
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import datetime
import logging
from typing import Callable, Dict, Iterable, List, Optional

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import mac

vollog = logging.getLogger(__name__)


class PsList(interfaces.plugins.PluginInterface):
    """Lists the processes present in a particular mac memory image."""

    _required_framework_version = (2, 0, 0)
    _version = (3, 0, 0)
    pslist_methods = ["tasks", "allproc", "process_group", "sessions", "pid_hash_table"]

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="macutils", component=mac.MacUtilities, version=(1, 1, 0)
            ),
            requirements.ChoiceRequirement(
                name="pslist_method",
                description="Method to determine for processes",
                choices=cls.pslist_methods,
                default=cls.pslist_methods[0],
                optional=True,
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    @classmethod
    def get_list_tasks(
        cls, method: str
    ) -> Callable[
        [interfaces.context.ContextInterface, str, Callable[[int], bool]],
        Iterable[interfaces.objects.ObjectInterface],
    ]:
        """Returns the list_tasks method based on the selector

        Args:
            method: Must be one of the available methods in get_task_choices

        Returns:
            list_tasks method for listing tasks
        """
        # Ensure method is one of the suitable choices
        if method not in cls.pslist_methods:
            method = cls.pslist_methods[0]

        if method == "allproc":
            list_tasks = cls.list_tasks_allproc
        elif method == "tasks":
            list_tasks = cls.list_tasks_tasks
        elif method == "process_group":
            list_tasks = cls.list_tasks_process_group
        elif method == "sessions":
            list_tasks = cls.list_tasks_sessions
        elif method == "pid_hash_table":
            list_tasks = cls.list_tasks_pid_hash_table
        else:
            raise ValueError("Impossible method choice chosen")
        vollog.debug(f"Using method {method}")

        return list_tasks

    @classmethod
    def create_pid_filter(
        cls, pid_list: Optional[List[int]] = None
    ) -> Callable[[int], bool]:
        def filter_func(_):
            return False

        # FIXME: mypy #4973 or #2608
        pid_list = pid_list or []
        filter_list = [x for x in pid_list if x is not None]
        if filter_list:

            def list_filter(x):
                return x.p_pid not in filter_list

            filter_func = list_filter
        return filter_func

    def _generator(self):
        list_tasks = self.get_list_tasks(
            self.config.get("pslist_method", self.pslist_methods[0])
        )

        for task in list_tasks(
            self.context,
            self.config["kernel"],
            filter_func=self.create_pid_filter(self.config.get("pid", None)),
        ):
            offset = format_hints.Hex(task.vol.offset)
            name = utility.array_to_string(task.p_comm)
            pid = task.p_pid
            uid = task.p_uid
            gid = task.p_gid
            start_time_seconds = task.p_start.tv_sec
            start_time_microseconds = task.p_start.tv_usec
            start_time = datetime.datetime.fromtimestamp(
                start_time_seconds + start_time_microseconds / 1e6
            )

            ppid = task.p_ppid

            yield (0, (offset, name, pid, uid, gid, start_time, ppid))

    @classmethod
    def list_tasks_allproc(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the processes in the primary layer based on the allproc method

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a process object and returns True if the process should be ignored/filtered

        Returns:
            The list of process objects from the processes linked list after filtering
        """

        kernel = context.modules[kernel_module_name]

        kernel_layer = context.layers[kernel.layer_name]

        proc = kernel.object_from_symbol(symbol_name="allproc").lh_first

        seen: Dict[int, int] = {}
        while proc is not None and proc.vol.offset != 0:
            if proc.vol.offset in seen:
                vollog.log(
                    logging.INFO,
                    "Recursive process list detected (a result of non-atomic acquisition).",
                )
                break
            else:
                seen[proc.vol.offset] = 1

            if kernel_layer.is_valid(
                proc.vol.offset, proc.vol.size
            ) and not filter_func(proc):
                yield proc

            try:
                proc = proc.p_list.le_next.dereference()
            except exceptions.InvalidAddressException:
                break

    @classmethod
    def list_tasks_tasks(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the tasks in the primary layer based on the tasks queue

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a task object and returns True if the task should be ignored/filtered

        Returns:
            The list of task objects from the `layer_name` layer's `tasks` list after filtering
        """
        kernel = context.modules[kernel_module_name]

        kernel_layer = context.layers[kernel.layer_name]

        queue_entry = kernel.object_from_symbol(symbol_name="tasks")

        seen: Dict[int, int] = {}
        for task in queue_entry.walk_list(queue_entry, "tasks", "task"):
            if task.vol.offset in seen:
                vollog.log(
                    logging.INFO,
                    "Recursive process list detected (a result of non-atomic acquisition).",
                )
                break
            else:
                seen[task.vol.offset] = 1

            try:
                proc = task.bsd_info.dereference().cast("proc")
            except exceptions.InvalidAddressException:
                continue

            if kernel_layer.is_valid(
                proc.vol.offset, proc.vol.size
            ) and not filter_func(proc):
                yield proc

    @classmethod
    def list_tasks_sessions(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the tasks in the primary layer using sessions

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a task object and returns True if the task should be ignored/filtered

        Returns:
            The list of task objects from the `layer_name` layer's `tasks` list after filtering
        """
        kernel = context.modules[kernel_module_name]

        table_size = kernel.object_from_symbol(symbol_name="sesshash")

        sesshashtbl = kernel.object_from_symbol(symbol_name="sesshashtbl")

        proc_array = kernel.object(
            object_type="array",
            offset=sesshashtbl,
            count=table_size + 1,
            subtype=kernel.get_type("sesshashhead"),
        )

        for proc_list in proc_array:
            for proc in mac.MacUtilities.walk_list_head(proc_list, "s_hash"):
                if proc.s_leader.is_readable() and not filter_func(proc.s_leader):
                    yield proc.s_leader

    @classmethod
    def list_tasks_process_group(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the tasks in the primary layer using process groups

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a task object and returns True if the task should be ignored/filtered

        Returns:
            The list of task objects from the `layer_name` layer's `tasks` list after filtering
        """
        kernel = context.modules[kernel_module_name]

        table_size = kernel.object_from_symbol(symbol_name="pgrphash")

        pgrphashtbl = kernel.object_from_symbol(symbol_name="pgrphashtbl")

        proc_array = kernel.object(
            object_type="array",
            offset=pgrphashtbl,
            count=table_size + 1,
            subtype=kernel.get_type("pgrphashhead"),
        )

        for proc_list in proc_array:
            for pgrp in mac.MacUtilities.walk_list_head(proc_list, "pg_hash"):
                for proc in mac.MacUtilities.walk_list_head(
                    pgrp.pg_members, "p_pglist"
                ):
                    if not filter_func(proc):
                        yield proc

    @classmethod
    def list_tasks_pid_hash_table(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the tasks in the primary layer using the pid hash table

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate
            filter_func: A function which takes a task object and returns True if the task should be ignored/filtered

        Returns:
            The list of task objects from the `layer_name` layer's `tasks` list after filtering
        """

        kernel = context.modules[kernel_module_name]

        table_size = kernel.object_from_symbol(symbol_name="pidhash")

        pidhashtbl = kernel.object_from_symbol(symbol_name="pidhashtbl")

        proc_array = kernel.object(
            object_type="array",
            offset=pidhashtbl,
            count=table_size + 1,
            subtype=kernel.get_type("pidhashhead"),
        )

        for proc_list in proc_array:
            for proc in mac.MacUtilities.walk_list_head(proc_list, "p_hash"):
                if not filter_func(proc):
                    yield proc

    def run(self):
        return renderers.TreeGrid(
            [
                ("OFFSET", format_hints.Hex),
                ("NAME", str),
                ("PID", int),
                ("UID", int),
                ("GID", int),
                ("Start Time", datetime.datetime),
                ("PPID", int),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List, Optional

from volatility3.framework import constants, exceptions, renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class CmdLine(interfaces.plugins.PluginInterface):
    """Lists process command line arguments."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
        ]

    @classmethod
    def get_cmdline(
        cls, context: interfaces.context.ContextInterface, kernel_table_name: str, proc
    ) -> Optional[str]:
        """Extracts the cmdline from PEB

        Args:
            context: the context to operate upon
            kernel_table_name: the name for the symbol table containing the kernel's symbols
            proc: the process object

        Returns:
            A string with the command line
        """

        proc_layer_name = proc.add_process_layer()
        if not proc_layer_name:
            return None

        peb = context.object(
            kernel_table_name + constants.BANG + "_PEB",
            layer_name=proc_layer_name,
            offset=proc.Peb,
        )

        return peb.ProcessParameters.CommandLine.get_string()

    def _generator(self, procs):
        kernel = self.context.modules[self.config["kernel"]]

        for proc in procs:
            process_name = utility.array_to_string(proc.ImageFileName)
            proc_id = "Unknown"
            result_text = None

            try:
                proc_id = proc.UniqueProcessId
                result_text = self.get_cmdline(
                    self.context, kernel.symbol_table_name, proc
                )

            except exceptions.SwappedInvalidAddressException as exp:
                vollog.debug(
                    f"Required memory at {exp.invalid_address:#x} is inaccessible (swapped)"
                )

            except exceptions.PagedInvalidAddressException as exp:
                vollog.debug(
                    f"Required memory at {exp.invalid_address:#x} is not valid (process exited?)"
                )

            except exceptions.InvalidAddressException as exp:
                vollog.debug(
                    f"Process {proc_id}: Required memory at {exp.invalid_address:#x} is not valid (incomplete layer {exp.layer_name}?)"
                )

            if not result_text:
                result_text = renderers.UnreadableValue()

            yield (0, (proc.UniqueProcessId, process_name, result_text))

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [("PID", int), ("Process", str), ("Args", str)],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from pathlib import PurePosixPath
from typing import Optional, Tuple, Iterator

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.constants import linux as linux_constants
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.symbols import linux
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


class ProcessSpoofing(plugins.PluginInterface):
    """Detects process spoofing by comparing executable path to cmdline & comm fields.

    Examples of such behavior can be found here: https://github.com/SolitudePy/linux-mal
    """

    _required_framework_version = (2, 27, 0)
    _version = (1, 1, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    @classmethod
    def get_executable_path(
        cls,
        context: interfaces.context.ContextInterface,
        task: interfaces.objects.ObjectInterface,
    ) -> Optional[str]:
        """
        Extract the executable path from task_struct.mm.exe_file

        Args:
            context: The context to operate on
            task: task_struct object of the process

        Returns:
            Returns executable path or None if not available
        """

        try:
            mm = task.mm
        except (exceptions.InvalidAddressException, AttributeError) as e:
            vollog.debug(f"Unable to access mm for task at {task.vol.offset:#x}: {e}")
            return None

        if not mm or not mm.is_readable():
            # Kernel threads don't have mm struct
            return None

        try:
            exe_file = mm.exe_file
        except (exceptions.InvalidAddressException, AttributeError) as e:
            vollog.debug(
                f"Unable to access exe_file for task at {task.vol.offset:#x}: {e}"
            )
            return None

        if not exe_file or not exe_file.is_readable():
            return None

        try:
            exe_path = linux.LinuxUtilities.path_for_file(context, task, exe_file)
        except (exceptions.InvalidAddressException, AttributeError) as e:
            vollog.debug(
                f"Unable to read exe_file path for task at {task.vol.offset:#x}: {e}"
            )
            return None

        return exe_path

    @classmethod
    def get_cmdline_basename(
        cls,
        context: interfaces.context.ContextInterface,
        task: interfaces.objects.ObjectInterface,
    ) -> Optional[str]:
        """
        Extract the command line arguments and return the basename of the first argument.

        Notes:
            The read length is capped at ``MAX_ARG_STRLEN`` (32 * 4096) per the
            kernel limit defined in ``include/uapi/linux/binfmts.h`` (see
            linux.git commit f6031913338f1dad5bd8cb7286ff4e53644b6940).

        Args:
            context: The context to operate on
            task: task_struct object of the process

        Returns:
            Basename of the first command line argument or None if not available
        """
        mm = task.mm
        if not mm or not mm.is_readable():
            return None

        proc_layer_name = task.add_process_layer()
        if proc_layer_name is None:
            return None

        start = task.mm.arg_start
        size_to_read = task.mm.arg_end - task.mm.arg_start

        if size_to_read <= 0:
            return None

        read_length = min(size_to_read, linux_constants.MAX_ARG_STRLEN)

        try:
            cmdline = utility.address_to_string(
                context=context,
                layer_name=proc_layer_name,
                address=start,
                count=read_length,
                errors="replace",
                encoding="utf-8",
            )
        except exceptions.InvalidAddressException as e:
            vollog.debug(
                f"Unable to read cmdline for task at {task.vol.offset:#x}: {e}"
            )
            return None

        if not cmdline:
            return None

        basename = PurePosixPath(cmdline).name
        return basename if basename else None

    @classmethod
    def get_comm(cls, task: interfaces.objects.ObjectInterface) -> Optional[str]:
        """
        Extract the comm field from task_struct

        Args:
            task: task_struct object of the process

        Returns:
            Process name from comm field or None if not available
        """
        try:
            return utility.array_to_string(task.comm)
        except (exceptions.InvalidAddressException, AttributeError) as e:
            vollog.debug(f"Unable to read comm for task at {task.vol.offset:#x}: {e}")
            return None

    @classmethod
    def extract_process_names(
        cls,
        context: interfaces.context.ContextInterface,
        task: interfaces.objects.ObjectInterface,
    ) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str], bool]:
        """
        Extract all process name sources for comparison

        Returns:
            Tuple of (exe_path, exe_basename, cmdline_basename, comm)
        """
        exe_path = cls.get_executable_path(context, task)
        exe_basename = PurePosixPath(exe_path).name if exe_path else None
        if exe_basename and exe_basename.endswith(" (deleted)"):
            exe_basename = exe_basename[: -len(" (deleted)")]
        cmdline_basename = cls.get_cmdline_basename(context, task)
        comm = cls.get_comm(task)

        return exe_path, exe_basename, cmdline_basename, comm

    def _detect_spoofing(
        self,
        exe_basename: Optional[str],
        cmdline_basename: Optional[str],
        comm: Optional[str],
    ) -> Tuple[bool, bool]:
        """
        Analyze the three name sources to detect potential spoofing

        Args:
            exe_basename: Basename from exe_file path
            cmdline_basename: Basename from command line
            comm: Name from comm field

        Returns:
            Tuple of (cmdline_spoofed, comm_spoofed) boolean flags
        """
        # Skip kernel threads - need at least 2 sources for comparison
        available_sources = sum(
            1 for name in [exe_basename, cmdline_basename, comm] if name
        )
        if available_sources < 2:
            return False, False

        # Check for cmdline spoofing
        cmdline_spoofed = False
        if exe_basename and cmdline_basename:
            cmdline_spoofed = exe_basename != cmdline_basename

        # Check for comm spoofing (comm is truncated to 15 characters)
        comm_spoofed = False
        if exe_basename and comm:
            comm_spoofed = exe_basename[:15] != comm

        return cmdline_spoofed, comm_spoofed

    def _generator(self, tasks) -> Iterator[Tuple[int, Tuple]]:
        """
        Generate process spoofing detection results

        Args:
            tasks: Iterator of task_struct objects

        Yields:
            Tuple containing process information and spoofing analysis
        """
        for task in tasks:
            try:
                pid = task.pid
                ppid = task.get_parent_pid()

                exe_path, exe_basename, cmdline_basename, comm = (
                    self.extract_process_names(self.context, task)
                )

                cmdline_spoofed, comm_spoofed = self._detect_spoofing(
                    exe_basename, cmdline_basename, comm
                )

                is_deleted = exe_path.endswith(" (deleted)") if exe_path else False

                # Convert None values to strings for TreeGrid compatibility
                exe_path_render = exe_path if exe_path else "N/A"
                cmdline_render = cmdline_basename if cmdline_basename else "N/A"
                comm_render = comm if comm else "N/A"

                yield (
                    0,
                    (
                        pid,
                        ppid,
                        exe_path_render,
                        cmdline_render,
                        comm_render,
                        cmdline_spoofed,
                        comm_spoofed,
                        is_deleted,
                    ),
                )

            except (exceptions.InvalidAddressException, AttributeError) as e:
                vollog.warning(
                    f"Unable to process task PID {getattr(task, 'pid', 'unknown')} at {task.vol.offset:#x}: {e}"
                )
                continue

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("PPID", int),
                ("Exe_Path", str),
                ("Cmdline_Basename", str),
                ("Comm", str),
                ("Cmdline_Spoofed", bool),
                ("Comm_Spoofed", bool),
                ("Exe_Deleted", bool),
            ],
            self._generator(
                pslist.PsList.list_tasks(
                    self.context, self.config["kernel"], filter_func=filter_func
                )
            ),
        )
